from copy import copy
from functools import lru_cache, partial
from typing import Set, Dict, Any, Optional, List, Union

from dlt.common.configuration import known_sections, resolve_configuration, with_config
from dlt.common import logger, json
from dlt.common.configuration.specs import BaseConfiguration, configspec
from dlt.common.destination.capabilities import (
    DestinationCapabilitiesContext,
    adjust_schema_to_capabilities,
)
from dlt.common.exceptions import MissingDependencyException
from dlt.common.metrics import DataWriterMetrics
from dlt.common.runtime.collector import Collector, NULL_COLLECTOR
from dlt.common.typing import TDataItems, TDataItem, TLoaderFileFormat
from dlt.common.schema import Schema, utils
from dlt.common.schema.typing import (
    C_DLT_LOAD_ID,
    TSchemaContractDict,
    TSchemaEvolutionMode,
    TTableSchema,
    TTableSchemaColumns,
    TPartialTableSchema,
)
from dlt.common.libs.sqlglot import filter_select_column_names
from dlt.common.normalizers.json import helpers as normalize_helpers

from dlt.extract.hints import HintsMeta, TResourceHints
from dlt.extract.resource import DltResource
from dlt.extract.items import DataItemWithMeta, TableNameMeta
from dlt.extract.storage import ExtractorItemStorage
from dlt.normalize.configuration import ItemsNormalizerConfiguration

try:
    from dlt.common.libs import pyarrow
    from dlt.common.libs.pyarrow import pyarrow as pa, TAnyArrowItem, UnsupportedArrowTypeException
except MissingDependencyException:
    pyarrow = None
    pa = None

try:
    from dlt.common.libs.pandas import pandas, pandas_to_arrow
except MissingDependencyException:
    pandas = None


class MaterializedEmptyList(List[Any]):
    """A list variant that will materialize tables even if empty list was yielded"""

    pass


def materialize_schema_item() -> MaterializedEmptyList:
    """Yield this to materialize schema in the destination, even if there's no data."""
    return MaterializedEmptyList()


class ImportFileMeta(HintsMeta):
    __slots__ = ("file_path", "metrics", "file_format")

    def __init__(
        self,
        file_path: str,
        metrics: DataWriterMetrics,
        file_format: TLoaderFileFormat = None,
        hints: TResourceHints = None,
        create_table_variant: bool = None,
    ) -> None:
        super().__init__(hints=hints, create_table_variant=create_table_variant)
        self.file_path = file_path
        self.metrics = metrics
        self.file_format = file_format


def with_file_import(
    file_path: str,
    file_format: TLoaderFileFormat,
    items_count: int = 0,
    hints: Union[TResourceHints, TDataItem] = None,
) -> DataItemWithMeta:
    """Marks file under `file_path` to be associated with current resource and imported into the load package as a file of
    type `file_format`.

    You can provide optional `hints` that will be applied to the current resource. Note that you should avoid schema inference at
    runtime if possible and if that is not possible - to do that only once per extract process. Use `make_hints` in `mark` module
    to create hints. You can also pass Arrow table or Pandas data frame form which schema will be taken (but content discarded).
    Create `TResourceHints` with `make_hints`.

    If number of records in `file_path` is known, pass it in `items_count` so `dlt` can generate correct extract metrics.

    Note that `dlt` does not sniff schemas from data and will not guess right file format for you.
    """
    metrics = DataWriterMetrics(file_path, items_count, 0, 0, 0)
    item: TDataItem = None
    # if hints are dict assume that this is dlt schema, if not - that it is arrow table
    if not isinstance(hints, dict):
        item = hints
        hints = None
    return DataItemWithMeta(ImportFileMeta(file_path, metrics, file_format, hints, False), item)


class Extractor:
    @configspec
    class ExtractorConfiguration(BaseConfiguration):
        _caps: Optional[DestinationCapabilitiesContext] = None

    @with_config(spec=ExtractorConfiguration)
    def __init__(
        self,
        load_id: str,
        item_storage: ExtractorItemStorage,
        schema: Schema,
        collector: Collector = NULL_COLLECTOR,
        *,
        _caps: DestinationCapabilitiesContext = None,
    ) -> None:
        self.schema = schema
        self.naming = schema.naming
        self.collector = collector
        self.resources_with_items: Set[str] = set()
        """Tracks resources that received items"""
        self.resources_with_empty: Set[str] = set()
        """Track resources that received empty materialized list"""
        self.load_id = load_id
        self.item_storage = item_storage
        self._table_contracts: Dict[str, TSchemaContractDict] = {}
        self._filtered_tables: Set[str] = set()
        self._filtered_columns: Dict[str, Dict[str, TSchemaEvolutionMode]] = {}
        self._caps = _caps or DestinationCapabilitiesContext.generic_capabilities()
        self._normalize_table_identifier = lru_cache(maxsize=None)(
            partial(normalize_helpers.normalize_table_identifier, self.schema, self.naming)
        )

    def write_items(self, resource: DltResource, items: TDataItems, meta: Any) -> None:
        """Write `items` to `resource` optionally computing table schemas and revalidating/filtering data"""

        if isinstance(meta, HintsMeta) and meta.hints:
            # update the resource with new hints, remove all caches so schema is recomputed
            # and contracts re-applied
            resource.merge_hints(meta.hints, meta.create_table_variant)
            # convert to table meta if created table variant so item is assigned to this table
            if meta.create_table_variant:
                # name in hints meta must be a string, otherwise merge_hints would fail
                meta = TableNameMeta(meta.hints["table_name"])  # type: ignore[arg-type]
            self._reset_contracts_cache()

        if resource.has_dynamic_table_name:
            # table has name or other hints depending on data items
            self._write_to_dynamic_table(resource, items, meta)
        else:
            # write item belonging to table with static name
            table_name = self._get_static_table_name(resource, meta)
            self._write_to_static_table(resource, table_name, items, meta)

    def write_empty_items_file(self, table_name: str) -> None:
        table_name = self._normalize_table_identifier(table_name)
        self.item_storage.write_empty_items_file(self.load_id, self.schema.name, table_name, None)

    def _get_static_table_name(self, resource: DltResource, meta: Any) -> str:
        if isinstance(meta, TableNameMeta):
            table_name = meta.table_name
        else:
            table_name = resource.table_name  # type: ignore[assignment]
        return self._normalize_table_identifier(table_name)

    def _get_dynamic_table_name(self, resource: DltResource, item: TDataItem) -> str:
        return (
            self._normalize_table_identifier(resource._table_name_hint_fun(item))
            if resource._table_name_hint_fun
            else self._get_static_table_name(resource, item)
        )

    def _write_item(
        self,
        table_name: str,
        resource_name: str,
        items: TDataItems,
        columns: TTableSchemaColumns = None,
    ) -> None:
        new_rows_count = self.item_storage.write_data_item(
            self.load_id, self.schema.name, table_name, items, columns
        )
        self.collector.update(table_name, inc=new_rows_count)
        # if there were rows or item was empty arrow table
        if new_rows_count > 0 or self.__class__ is ArrowExtractor:
            self.resources_with_items.add(resource_name)
        else:
            if isinstance(items, MaterializedEmptyList):
                self.resources_with_empty.add(resource_name)

    def _import_item(
        self,
        table_name: str,
        resource_name: str,
        meta: ImportFileMeta,
    ) -> None:
        metrics = self.item_storage.import_items_file(
            self.load_id,
            self.schema.name,
            table_name,
            meta.file_path,
            meta.metrics,
            meta.file_format,
        )
        self.collector.update(table_name, inc=metrics.items_count)
        self.resources_with_items.add(resource_name)

    def _write_to_dynamic_table(self, resource: DltResource, items: TDataItems, meta: Any) -> None:
        if not isinstance(items, list):
            items = [items]

        for item in items:
            table_name = self._get_dynamic_table_name(resource, item)
            # don't write table if contract says so
            if table_name in self._filtered_tables:
                continue
            # allow to cache dynamic table hints if only table name is dynamic
            if table_name not in self._table_contracts or resource.has_other_dynamic_hints:
                item = self._compute_and_update_tables(
                    resource, table_name, item, TableNameMeta(table_name)
                )
            # write to storage with inferred table name
            if table_name not in self._filtered_tables:
                if isinstance(meta, ImportFileMeta):
                    self._import_item(table_name, resource.name, meta)
                else:
                    self._write_item(table_name, resource.name, item)

    def _write_to_static_table(
        self, resource: DltResource, table_name: str, items: TDataItems, meta: Any
    ) -> None:
        if table_name not in self._table_contracts:
            items = self._compute_and_update_tables(resource, table_name, items, meta)
        if table_name not in self._filtered_tables:
            if isinstance(meta, ImportFileMeta):
                self._import_item(table_name, resource.name, meta)
            else:
                self._write_item(table_name, resource.name, items)

    def _compute_tables(
        self, resource: DltResource, items: TDataItems, meta: Any
    ) -> List[TTableSchema]:
        """Computes a schema for a new or dynamic table and normalizes identifiers"""
        root_table_schema = resource.compute_table_schema(items, meta)
        nested_tables_schema = resource.compute_nested_table_schemas(
            root_table_schema["name"], self.naming, items, meta
        )
        # we need to re-normalize name to support legacy normalization mode which we will
        # drop in next major version
        # TODO: drop in 2.0 also drop SCHEMA__USE_BREAK_PATH_ON_NORMALIZE
        root_table_schema["name"] = self._normalize_table_identifier(root_table_schema["name"])
        return [
            utils.normalize_table_identifiers(table_schema, self.naming)
            for table_schema in (root_table_schema, *nested_tables_schema)
        ]

    def _compute_and_update_tables(
        self, resource: DltResource, root_table_name: str, items: TDataItems, meta: Any
    ) -> TDataItems:
        """
        Computes new table and does contract checks, if false is returned, the table may not be created and no items should be written
        """
        computed_tables = self._compute_tables(resource, items, meta)
        # pre-merge authoritative schema from validator into self.schema
        # so authoritative columns bypass contract checks via diff
        if resource.validator:
            auth_schema = resource.validator.compute_table_schema(items, meta)
            if auth_schema:
                self.schema.update_table(
                    auth_schema, normalize_identifiers=True, merge_compound_props=False
                )
        for computed_table in computed_tables:
            table_name = computed_table["name"]
            # get or compute contract
            schema_contract = self._table_contracts.setdefault(
                table_name,
                self.schema.resolve_contract_settings_for_table(table_name, computed_table),
            )

            # this is a new table so allow evolve once
            if schema_contract["columns"] != "evolve" and self.schema.is_new_table(table_name):
                computed_table["x-normalizer"] = {"evolve-columns-once": True}

            # apply contracts
            computed_table, filters = self.schema.apply_schema_contract(
                schema_contract, computed_table, data_item=items
            )

            # merge with schema table
            if computed_table:
                # computed table identifiers already normalized
                self.schema.update_table(
                    computed_table,
                    normalize_identifiers=False,
                    merge_compound_props=False,
                )

            # process filters
            if filters:
                for entity, name, mode in filters:
                    if entity == "tables":
                        self._filtered_tables.add(name)
                    elif entity == "columns":
                        filtered_columns = self._filtered_columns.setdefault(table_name, {})
                        filtered_columns[name] = mode
        return items

    def _reset_contracts_cache(self) -> None:
        """Removes all cached contracts, filtered columns and tables"""
        self._table_contracts.clear()
        self._filtered_tables.clear()
        self._filtered_columns.clear()


class ObjectExtractor(Extractor):
    """Extracts Python object data items into typed jsonl"""

    pass


class ModelExtractor(Extractor):
    """Extracts model items (SQL Relations) with data_type contract enforcement."""

    def _compute_and_update_tables(
        self, resource: DltResource, root_table_name: str, items: TDataItems, meta: Any
    ) -> TDataItems:
        items = super()._compute_and_update_tables(resource, root_table_name, items, meta)
        return self._apply_contract_filters(items, root_table_name)

    def _apply_contract_filters(self, items: TDataItems, table_name: str) -> TDataItems:
        """Apply data_type contract filters to model items (Relations)."""
        filtered_columns = self._filtered_columns.get(table_name)
        if not filtered_columns:
            return items

        if any(mode == "discard_row" for mode in filtered_columns.values()):
            return items.limit(0)  # type: ignore[union-attr]

        discard_value_cols = {
            name for name, mode in filtered_columns.items() if mode == "discard_value"
        }
        if discard_value_cols:
            selects = items.sqlglot_expression.selects  # type: ignore[union-attr]
            remaining = filter_select_column_names(
                selects, discard_value_cols, self.naming.normalize_identifier
            )
            if len(remaining) < len(selects):
                items = items.select(*remaining)  # type: ignore[union-attr]

        return items


class ArrowExtractor(Extractor):
    """Extracts arrow data items into parquet. Normalizes arrow items column names.
    Compares the arrow schema to actual dlt table schema to reorder the columns and to
    insert missing columns (without data). Adds _dlt_load_id column to the table if
    `add_dlt_load_id` is set to True in normalizer config.

    We do things that normalizer should do here so we do not need to load and save parquet
    files again later.

    Handles the following types:
    - `pyarrow.Table`
    - `pyarrow.RecordBatch`
    - `pandas.DataFrame` (is converted to arrow `Table` before processing)
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._normalize_config = self._retrieve_normalize_config()

    def _retrieve_normalize_config(self) -> ItemsNormalizerConfiguration:
        """Get normalizer settings that are used here"""
        return resolve_configuration(
            ItemsNormalizerConfiguration(),
            sections=(known_sections.NORMALIZE, "parquet_normalizer"),
        )

    def write_items(self, resource: DltResource, items: TDataItems, meta: Any) -> None:
        items_list = items if isinstance(items, list) else [items]

        static_table_name = self._get_static_table_name(resource, meta)
        items = [
            # 2. remove columns and rows in data contract filters
            self._apply_contract_filters(tbl, resource, static_table_name)
            for tbl in (
                (
                    # 1. Convert pandas frame(s) to arrow Table, remove indexes because we store
                    pandas_to_arrow(item)
                    if (pandas and isinstance(item, pandas.DataFrame))
                    else item
                )
                for item in items_list
            )
        ]
        super().write_items(resource, items, meta)

    def _write_to_static_table(
        self, resource: DltResource, table_name: str, items: TDataItems, meta: Any
    ) -> None:
        # contract cache not supported for arrow tables
        self._reset_contracts_cache()
        super()._write_to_static_table(resource, table_name, items, meta)

    def _apply_contract_filters(
        self, item: "TAnyArrowItem", resource: DltResource, static_table_name: Optional[str]
    ) -> "TAnyArrowItem":
        """Removes the columns (discard value) or rows (discard rows) as indicated by contract filters."""
        # convert arrow schema names into normalized names
        rename_mapping = pyarrow.get_normalized_arrow_fields_mapping(item.schema, self.naming)
        # find matching columns and delete by original name
        table_name = static_table_name or self._get_dynamic_table_name(resource, item)
        filtered_columns = self._filtered_columns.get(table_name)
        if filtered_columns:
            # remove rows where columns have non null values
            # create a mask where rows will be False if any of the specified columns are non-null
            mask = None
            rev_mapping = {v: k for k, v in rename_mapping.items()}
            for column in [
                name for name, mode in filtered_columns.items() if mode == "discard_row"
            ]:
                is_null = pyarrow.pyarrow.compute.is_null(item[rev_mapping[column]])
                mask = is_null if mask is None else pyarrow.pyarrow.compute.and_(mask, is_null)
            # filter the table using the mask
            if mask is not None:
                item = item.filter(mask)

            # remove value actually removes the whole columns from the table
            # NOTE: filtered columns has normalized column names so we need to go through mapping
            removed_columns = [
                name
                for name in rename_mapping
                if filtered_columns.get(rename_mapping[name]) is not None
            ]
            if removed_columns:
                item = pyarrow.remove_columns(item, removed_columns)

        return item

    def _write_item(
        self,
        table_name: str,
        resource_name: str,
        items: TDataItems,
        columns: TTableSchemaColumns = None,
    ) -> None:
        columns = columns or self.schema.get_table_columns(table_name)
        # Note: `items` is always a list here due to the conversion in `write_table`
        items = [
            pyarrow.normalize_py_arrow_item(
                item,
                columns,
                self.naming,
                self._caps,
            )
            for item in items
        ]

        if self._normalize_config.add_dlt_load_id:
            items = [
                pyarrow.add_dlt_load_id_column(
                    item,
                    columns,
                    self._caps,
                    self.naming,
                    self.load_id,
                )
                for item in items
            ]

        # write items one by one
        super()._write_item(table_name, resource_name, items, columns)

    def _compute_tables(
        self, resource: DltResource, items: TDataItems, meta: Any
    ) -> List[TPartialTableSchema]:
        arrow_tables: Dict[str, TTableSchema] = {}

        if isinstance(items, list):
            items = reversed(items)
        else:
            items = [items]

        # several arrow tables will update the pipeline schema and we want that earlier
        # arrow tables override the latter so the resultant schema is the same as if
        # they are sent separately
        for item in items:
            computed_tables = super()._compute_tables(resource, item, meta)
            for computed_table in computed_tables:
                arrow_table = arrow_tables.get(computed_table["name"])
                # Merge the columns to include primary_key and other hints that may be set on the resource
                if arrow_table:
                    utils.merge_table(self.schema.name, computed_table, arrow_table)
                else:
                    arrow_table = copy(computed_table)
                try:
                    # generate dlt schema from arrow schema and adjust to capabilities
                    # drop timezones and honor only explicit settings like the regular normalizer
                    arrow_table["columns"] = adjust_schema_to_capabilities(
                        pyarrow.py_arrow_to_table_schema_columns(item.schema),
                        self._caps,
                    )
                    # drop incomplete columns inferred from pa.null() arrow fields
                    arrow_table["columns"] = {
                        k: v
                        for k, v in arrow_table["columns"].items()
                        if utils.is_complete_column(v)
                    }
                except pyarrow.UnsupportedArrowTypeException as e:
                    e.table_name = str(arrow_table.get("name"))
                    raise

                # normalize arrow table before merging
                arrow_table = utils.normalize_table_identifiers(arrow_table, self.schema.naming)

                # Add load_id column if needed
                dlt_load_id = self.naming.normalize_identifier(C_DLT_LOAD_ID)
                if (
                    self._normalize_config.add_dlt_load_id
                    and dlt_load_id not in arrow_table["columns"]
                ):
                    # will be normalized line below
                    logger.debug(
                        f"Arrow Extractor added `{dlt_load_id}` to table `{arrow_table['name']}`"
                        " due to parquet normalizer config"
                    )
                    arrow_table["columns"][dlt_load_id] = utils.dlt_load_id_column()

                # issue warnings when overriding computed with arrow
                override_warn: bool = False
                for col_name, column in arrow_table["columns"].items():
                    if src_column := computed_table["columns"].get(col_name):
                        for hint_name, hint in column.items():
                            if (src_hint := src_column.get(hint_name)) is not None:
                                if src_hint != hint:
                                    override_warn = True
                                    logger.info(
                                        f"In resource: {resource.name}, when merging arrow schema"
                                        f" on column {col_name}. The hint {hint_name} value"
                                        f" {src_hint} defined in resource will overwrite arrow hint"
                                        f" with value {hint}."
                                    )
                if override_warn:
                    logger.warning(
                        f"In resource: {resource.name}, when merging arrow schema with dlt schema,"
                        " several column hints were different. dlt schema hints were kept and arrow"
                        " schema and data were unmodified. It is up to destination to coerce the"
                        " differences when loading. Change log level to INFO for more details."
                    )
                utils.merge_columns(arrow_table["columns"], computed_table["columns"])
                arrow_tables[computed_table["name"]] = arrow_table

        return list(arrow_tables.values())

    def _compute_and_update_tables(
        self, resource: DltResource, root_table_name: str, items: TDataItems, meta: Any
    ) -> TDataItems:
        items = super()._compute_and_update_tables(resource, root_table_name, items, meta)

        if not isinstance(items, list):
            # filter data item as filters could be updated in compute table
            items = [items]
        items = [self._apply_contract_filters(item, resource, root_table_name) for item in items]
        return items



from typing import Any, Optional, Type, Union, Dict, TYPE_CHECKING

from dlt.common import logger
from dlt.common.destination import Destination, DestinationCapabilitiesContext
from dlt.common.destination.configuration import ParquetFormatConfiguration
from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.exceptions import TerminalValueError
from dlt.common.normalizers.naming.naming import NamingConvention
from dlt.common.data_writers.escape import escape_postgres_identifier, escape_mssql_literal
from dlt.common.arithmetics import DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE

from dlt.common.schema.typing import TColumnSchema, TColumnType
from dlt.destinations._adbc_jobs import make_adbc_parquet_file_format_selector
from dlt.destinations.type_mapping import TypeMapperImpl
from dlt.destinations.impl.mssql.configuration import MsSqlCredentials, MsSqlClientConfiguration

if TYPE_CHECKING:
    from dlt.destinations.impl.mssql.mssql import MsSqlJobClient


class MsSqlTypeMapper(TypeMapperImpl):
    sct_to_unbound_dbt = {
        "json": "json",
        "text": "nvarchar(max)",
        "double": "float",
        "bool": "bit",
        "bigint": "bigint",
        "binary": "varbinary(max)",
        "date": "date",
        "timestamp": "datetimeoffset",
        "time": "time",
    }

    sct_to_dbt = {
        "text": "nvarchar(%i)",
        "timestamp": "datetimeoffset(%i)",
        "binary": "varbinary(%i)",
        "decimal": "decimal(%i,%i)",
        "time": "time(%i)",
        "wei": "decimal(%i,%i)",
    }

    dbt_to_sct = {
        "nvarchar": "text",
        "float": "double",
        "bit": "bool",
        "datetimeoffset": "timestamp",
        "datetime2": "timestamp",
        "date": "date",
        "bigint": "bigint",
        "varbinary": "binary",
        "decimal": "decimal",
        "time": "time",
        "tinyint": "bigint",
        "smallint": "bigint",
        "int": "bigint",
        "json": "json",
    }

    def to_db_datetime_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema = None,
    ) -> str:
        column_name = column["name"]
        table_name = table["name"]
        timezone = column.get("timezone", True)
        precision = column.get("precision")

        # Synapse DATETIME2 allows precision from 0-7
        if precision is not None and (precision < 0 or precision > 7):
            logger.warn(
                "Azure Synapse/MSSQL only supports precision between 0-7 for column"
                f" '{column_name}' in table '{table_name}'. Will use precision 7."
            )
            precision = self.capabilities.max_timestamp_precision

        # NOTE: default precision on MSSQL is 7 and we'll use it. default caps precision is 6
        #  but setting it to 7 will generate nano precision timestamps in arrow
        precision_str = ""
        if precision is not None:
            precision_str = f"({precision})"

        if timezone:
            return f"datetimeoffset{precision_str}"  # mssql's timezone-aware datetime type
        else:
            return f"datetime2{precision_str}"  # mssql's high-precision datetime without timezone

    def to_db_integer_type(self, column: TColumnSchema, table: PreparedTableSchema = None) -> str:
        precision = column.get("precision")
        if precision is None:
            return "bigint"
        if precision <= 8:
            return "tinyint"
        if precision <= 16:
            return "smallint"
        if precision <= 32:
            return "int"
        elif precision <= 64:
            return "bigint"
        raise TerminalValueError(
            f"bigint with `{precision=:}` can't be mapped to MSSQL integer type"
        )

    def from_destination_type(
        self, db_type: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        if db_type == "decimal":
            if (precision, scale) == self.capabilities.wei_precision:
                return dict(data_type="wei")
        if db_type == "datetime2":
            return {"data_type": "timestamp", "timezone": False}
        return super().from_destination_type(db_type, precision, scale)


class mssql(Destination[MsSqlClientConfiguration, "MsSqlJobClient"]):
    spec = MsSqlClientConfiguration

    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        caps = DestinationCapabilitiesContext()
        caps.preferred_loader_file_format = "insert_values"
        caps.supported_loader_file_formats = ["insert_values", "parquet", "model"]
        caps.loader_file_format_selector = make_adbc_parquet_file_format_selector(
            "mssql",
            "https://dlthub.com/docs/dlt-ecosystem/destinations/mssql#data-loading",
            prefer_parquet=True,
        )
        caps.preferred_staging_file_format = None
        caps.supported_staging_file_formats = []
        caps.type_mapper = MsSqlTypeMapper
        # mssql is by default case insensitive and stores identifiers as is
        # case sensitivity can be changed by database collation so we allow to reconfigure
        # capabilities in the mssql factory
        caps.escape_identifier = escape_postgres_identifier
        caps.escape_literal = escape_mssql_literal
        caps.has_case_sensitive_identifiers = False
        caps.decimal_precision = (DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE)
        caps.wei_precision = (DEFAULT_NUMERIC_PRECISION, 0)
        # https://learn.microsoft.com/en-us/sql/sql-server/maximum-capacity-specifications-for-sql-server?view=sql-server-ver16&redirectedfrom=MSDN
        caps.max_identifier_length = 128
        caps.max_column_identifier_length = 128
        # A SQL Query can be a varchar(max) but is shown as limited to 65,536 * Network Packet
        caps.max_query_length = 65536 * 10
        caps.is_max_query_length_in_bytes = True
        caps.max_text_data_type_length = 2**30 - 1
        caps.is_max_text_data_type_length_in_bytes = False
        caps.supports_ddl_transactions = True
        caps.supports_multiple_statements = True
        caps.supports_create_table_if_not_exists = False  # IF NOT EXISTS not supported
        caps.max_rows_per_insert = 1000
        # NOTE: timestamp_precision is 7 in the database but there's no way to write it via Python
        caps.timestamp_precision = 6
        caps.max_timestamp_precision = 7
        caps.supported_merge_strategies = ["delete-insert", "upsert", "scd2"]
        caps.supported_replace_strategies = [
            "truncate-and-insert",
            "insert-from-staging",
            "staging-optimized",
        ]
        caps.sqlglot_dialect = "tsql"
        # ADBC driver for MSSQL does not support dictionary-encoded Arrow arrays
        caps.parquet_format = ParquetFormatConfiguration(supports_dictionary_encoding=False)

        return caps

    @property
    def client_class(self) -> Type["MsSqlJobClient"]:
        from dlt.destinations.impl.mssql.mssql import MsSqlJobClient

        return MsSqlJobClient

    def __init__(
        self,
        credentials: Union[MsSqlCredentials, Dict[str, Any], str] = None,
        create_indexes: bool = False,
        has_case_sensitive_identifiers: bool = False,
        destination_name: str = None,
        environment: str = None,
        **kwargs: Any,
    ) -> None:
        """Configure the MsSql destination to use in a pipeline.

        All arguments provided here supersede other configuration sources such as environment variables and dlt config files.

        Args:
            credentials (Union[MsSqlCredentials, Dict[str, Any], str], optional): Credentials to connect to the mssql database. Can be an instance of `MsSqlCredentials` or
                a connection string in the format `mssql://user:password@host:port/database`
            create_indexes (bool, optional): Should unique indexes be created
            has_case_sensitive_identifiers (bool, optional): Are identifiers used by mssql database case sensitive (following the collation)
            destination_name (str, optional): Name of the destination, can be used in config section to differentiate between multiple of the same type
            environment (str, optional): Environment of the destination
            **kwargs (Any): Additional arguments passed to the destination config
        """
        super().__init__(
            credentials=credentials,
            create_indexes=create_indexes,
            has_case_sensitive_identifiers=has_case_sensitive_identifiers,
            destination_name=destination_name,
            environment=environment,
            **kwargs,
        )

    @classmethod
    def adjust_capabilities(
        cls,
        caps: DestinationCapabilitiesContext,
        config: MsSqlClientConfiguration,
        naming: Optional[NamingConvention],
    ) -> DestinationCapabilitiesContext:
        # modify the caps if case sensitive identifiers are requested
        if config.has_case_sensitive_identifiers:
            caps.has_case_sensitive_identifiers = True
            caps.casefold_identifier = str
        return super().adjust_capabilities(caps, config, naming)


mssql.register()



import re
from typing import TYPE_CHECKING, Any, BinaryIO, IO
import os
from pathlib import Path
import sys
import base64
import hashlib
import secrets
from contextlib import contextmanager
from functools import wraps
from os import environ
from types import ModuleType
import traceback
import zlib
import tarfile
from importlib.metadata import version as pkg_version
from packaging.version import Version

if TYPE_CHECKING:
    from dlt.common.libs.sqlglot import TSqlGlotDialect
else:
    TSqlGlotDialect = Any

from typing import (
    Any,
    Callable,
    ContextManager,
    Dict,
    MutableMapping,
    Iterator,
    Generator,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Mapping,
    List,
    Union,
    Iterable,
    IO,
    cast,
)

from dlt.common.exceptions import (
    DltException,
    ExceptionTrace,
    TerminalException,
    DependencyVersionException,
    ValueErrorWithKnownValues,
)
from dlt.common.typing import AnyFun, StrAny, DictStrAny, StrStr, TAny, TFun, Generic
from dlt.common.warnings import Dlt100DeprecationWarning, deprecated


T = TypeVar("T")
TObj = TypeVar("TObj", bound=object)
TDict = TypeVar("TDict", bound=MutableMapping[Any, Any])

TKey = TypeVar("TKey")
TValue = TypeVar("TValue")

# row counts
RowCounts = Dict[str, int]


def chunks(iterable: Iterable[T], n: int) -> Iterator[Sequence[T]]:
    it = iter(iterable)
    while True:
        chunk = list()
        try:
            for _ in range(n):
                chunk.append(next(it))
        except StopIteration:
            if chunk:
                yield chunk
            break
        yield chunk


def uniq_id(len_: int = 16) -> str:
    """Returns a hex encoded crypto-grade string of random bytes with desired len_"""
    return secrets.token_hex(len_)


def uniq_id_base64(len_: int = 16) -> str:
    """Returns a base64 encoded crypto-grade string of random bytes with desired len_"""
    return base64.b64encode(secrets.token_bytes(len_)).decode("ascii").rstrip("=")


def many_uniq_ids_base64(n_ids: int, len_: int = 16) -> List[str]:
    """Generate `n_ids` base64 encoded crypto-grade strings of random bytes with desired len_.
    This is more performant than calling `uniq_id_base64` multiple times.
    """
    random_bytes = secrets.token_bytes(n_ids * len_)
    encode = base64.b64encode
    return [
        encode(random_bytes[i : i + len_]).decode("ascii").rstrip("=")
        for i in range(0, n_ids * len_, len_)
    ]


def digest128(v: str, len_: int = 15) -> str:
    """Returns a base64 encoded shake128 hash of str `v` with digest of length `len_` (default: 15 bytes = 20 characters length)"""
    return (
        base64.b64encode(hashlib.shake_128(v.encode("utf-8")).digest(len_))
        .decode("ascii")
        .rstrip("=")
    )


def digest128b(v: bytes, len_: int = 15) -> str:
    """Returns a base64 encoded shake128 hash of bytes `v` with digest of length `len_` (default: 15 bytes = 20 characters length)"""
    enc_v = base64.b64encode(hashlib.shake_128(v).digest(len_)).decode("ascii")
    return enc_v.rstrip("=")


def digest256(v: str) -> str:
    digest = hashlib.sha3_256(v.encode("utf-8")).digest()
    return base64.b64encode(digest).decode("ascii")


def digest256_file_stream(stream: BinaryIO, chunk_size: int = 4096) -> str:
    """Returns a base64 encoded sha3_256 hash of a binary stream"""
    stream.seek(0)
    hash_obj = hashlib.sha3_256()
    while chunk := stream.read(chunk_size):
        hash_obj.update(chunk)
    digest = hash_obj.digest()
    return base64.b64encode(digest).decode("ascii")


def digest256_tar_stream(
    stream: IO[bytes],
    filter_file_names: Callable[[str], bool] = lambda x: True,
    chunk_size: int = 8192,
) -> Tuple[str, List[str]]:
    """Calculates hash and collects file names from tar archive in a single pass.

    Hashes only file names and file contents of filtered members, ignoring timestamps
    and other tar metadata. Members are sorted by name before hashing for consistency.
    Operates entirely in-memory to prevent leakage of sensitive data.

    Args:
        stream: Binary stream containing the tar archive
        filter_file_names: Callable that returns True for members to include in hash
            and file names list. Default includes all members. Use this to exclude
            metadata files (e.g., manifest.yaml) from the hash calculation.
        chunk_size: Size of chunks to read when hashing file contents. Default 8192.

    Returns:
        tuple: (content_hash, file_names)
    """
    stream.seek(0)
    hash_obj = hashlib.sha3_256()
    file_names = []

    with tarfile.open(fileobj=stream, mode="r:*") as tar:
        members = sorted(tar.getmembers(), key=lambda m: m.name)

        for member in members:
            if not filter_file_names(member.name):
                continue

            hash_obj.update(member.name.encode())
            if member.isfile():
                file_names.append(member.name)
                f = tar.extractfile(member)
                if f:
                    while chunk := f.read(chunk_size):
                        hash_obj.update(chunk)

    digest = hash_obj.digest()
    content_hash = base64.b64encode(digest).decode("ascii")

    return content_hash, file_names


def str2bool(v: str) -> bool:
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise ValueErrorWithKnownValues(
            "v", v, [True, "yes", "true", "t", "y", 1, False, "no", "false", "f", "n", "0"]
        )


# def flatten_list_of_dicts(dicts: Sequence[StrAny]) -> StrAny:
#     """
#     Transforms a list of objects [{K: {...}}, {L: {....}}, ...] -> {K: {...}, L: {...}...}
#     """
#     o: DictStrAny = {}
#     for d in dicts:
#         for k,v in d.items():
#             if k in o:
#                 raise KeyError(f"Cannot flatten with duplicate key {k}")
#             o[k] = v
#     return o


def flatten_list_of_str_or_dicts(seq: Sequence[Union[StrAny, str]]) -> DictStrAny:
    """
    Transforms a list of objects or strings [{K: {...}}, L, ...] -> {K: {...}, L: None, ...}
    """
    o: DictStrAny = {}
    for e in seq:
        if isinstance(e, dict):
            for k, v in e.items():
                if k in o:
                    raise KeyError(f"Failed to flatten because of duplicate key `{k}`")
                o[k] = v
        else:
            key = str(e)
            if key in o:
                raise KeyError(f"Failed to flatten because of duplicate key `{key}`")
            o[key] = None
    return o


def flatten_list_or_items(
    _iter: Union[Iterable[TAny], Iterable[List[TAny]]]
) -> Generator[TAny, None, None]:
    for items in _iter:
        if isinstance(items, List):
            yield from items
        else:
            yield items


def concat_strings_with_limit(strings: List[str], separator: str, limit: int) -> Iterator[str]:
    """
    Generator function to concatenate strings.

    The function takes a list of strings and concatenates them into a single string such that the length of each
    concatenated string does not exceed a specified limit. It yields each concatenated string as it is created.
    The strings are separated by a specified separator.

    Args:
        strings (List[str]): The list of strings to be concatenated.
        separator (str): The separator to use between strings. Defaults to a single space.
        limit (int): The maximum length for each concatenated string.

    Yields:
        Generator[str, None, None]: A generator that yields each concatenated string.
    """

    if not strings:
        return

    current_length = len(strings[0])
    start = 0
    sep_len = len(separator)

    for i in range(1, len(strings)):
        if (
            current_length + len(strings[i]) + sep_len > limit
        ):  # accounts for the length of separator
            yield separator.join(strings[start:i])
            start = i
            current_length = len(strings[i])
        else:
            current_length += len(strings[i]) + sep_len  # accounts for the length of separator

    yield separator.join(strings[start:])


def graph_edges_to_nodes(
    edges: Sequence[Tuple[TAny, TAny]], directed: bool = True
) -> Dict[TAny, Set[TAny]]:
    """Converts a directed graph represented as a sequence of edges to a graph represented as a mapping from nodes a set of connected nodes.

    Isolated nodes are represented as edges to itself. If `directed` is `False`, each edge is duplicated but going in opposite direction.
    """
    graph: Dict[TAny, Set[TAny]] = {}
    for u, v in edges:
        if u not in graph:
            graph[u] = set()
        if v not in graph:
            graph[v] = set()
        if v != u:
            graph[u].add(v)
            if not directed:
                graph[v].add(u)

    return graph


def graph_find_scc_nodes(undag: Dict[TAny, Set[TAny]]) -> List[Set[TAny]]:
    """Finds and returns a list of sets of nodes in strongly connected components of a `undag` which is undirected

    To obtain undirected graph from edges use `graph_edges_to_nodes` function with `directed` argument `False`.
    """
    visited: Set[TAny] = set()
    components: List[Set[TAny]] = []

    def dfs(node: TAny, current_component: Set[TAny]) -> None:
        if node not in visited:
            visited.add(node)
            current_component.add(node)
            for neighbor in undag[node]:
                dfs(neighbor, current_component)

    for node in undag:
        if node not in visited:
            component: Set[TAny] = set()
            dfs(node, component)
            components.append(component)

    return components


def filter_env_vars(envs: List[str]) -> StrStr:
    return {k.lower(): environ[k] for k in envs if k in environ}


def update_dict_with_prune(dest: DictStrAny, update: StrAny) -> None:
    """Updates values that are both in `dest` and `update` and deletes `dest` values that are None in `update`"""
    for k, v in update.items():
        if v is not None:
            dest[k] = v
        elif k in dest:
            del dest[k]


def update_dict_nested(dst: TDict, src: TDict, copy_src_dicts: bool = False) -> TDict:
    """Merges `src` into `dst` key wise. Does not recur into lists. Values in `src` overwrite `dst` if both keys exit.
    Only `dict` and its subclasses are updated recursively. With `copy_src_dicts`, dict key:values will be deep copied,
    otherwise, both dst and src will keep the same references.
    """

    for key in src:
        src_val = src[key]
        if key in dst:
            dst_val = dst[key]
            if isinstance(src_val, dict) and isinstance(dst_val, dict):
                # If the key for both `dst` and `src` are both Mapping types (e.g. dict), then recurse.
                update_dict_nested(dst_val, src_val, copy_src_dicts=copy_src_dicts)
                continue

        if copy_src_dicts and isinstance(src_val, dict):
            dst[key] = update_dict_nested({}, src_val, True)
        else:
            dst[key] = src_val
    return dst


def clone_dict_nested(src: TDict) -> TDict:
    """Clones `src` structure descending into nested dicts. Does not descend into mappings that are not dicts ie. specs instances.
    Compared to `deepcopy` does not clone any other objects. Uses `update_dict_nested` internally
    """
    return update_dict_nested({}, src, copy_src_dicts=True)  # type: ignore[return-value]


def map_nested_values_in_place(
    func: AnyFun,
    _nested: TAny,
    *args: Any,
    **kwargs: Any,
) -> TAny:
    """Applies `func` to all values in `_dict` recursively, replacing elements in nested dictionaries and lists in place.
    Additional `*args` and `**kwargs` are passed to `func`.
    """
    if isinstance(_nested, tuple):
        if hasattr(_nested, "_asdict"):
            _nested = _nested._asdict()
        else:
            _nested = list(_nested)  # type: ignore

    if isinstance(_nested, dict):
        for k, v in _nested.items():
            if isinstance(v, (dict, list, tuple)):
                _nested[k] = map_nested_values_in_place(func, v, *args, **kwargs)
            else:
                _nested[k] = func(v, *args, **kwargs)
    elif isinstance(_nested, list):
        for idx, _l in enumerate(_nested):
            if isinstance(_l, (dict, list, tuple)):
                _nested[idx] = map_nested_values_in_place(func, _l, *args, **kwargs)
            else:
                _nested[idx] = func(_l, *args, **kwargs)
    else:
        raise ValueError(_nested, "Not a nested type")
    return _nested


# keep old name for backwards compatibility
map_nested_in_place = deprecated(
    "Use `map_nested_values_in_place` instead.", category=Dlt100DeprecationWarning
)(map_nested_values_in_place)


def map_nested_keys_in_place(
    func: AnyFun,
    _nested: TAny,
    *args: Any,
    **kwargs: Any,
) -> TAny:
    """Applies `func` to all keys in `_dict` recursively, replacing elements in nested dictionaries and lists in place.
    Additional `*args` and `**kwargs` are passed to `func`. List indexes will remain untouched.
    """
    if isinstance(_nested, tuple):
        if hasattr(_nested, "_asdict"):
            _nested = _nested._asdict()
        else:
            _nested = list(_nested)  # type: ignore

    if isinstance(_nested, dict):
        # NOTE: to modify the dictionary in place, exhaust the iterator into a list before iterating over it
        for k, v in list(_nested.items()):
            _nested.pop(k)
            k = func(k, *args, **kwargs)
            if isinstance(v, (dict, list, tuple)):
                _nested[k] = map_nested_keys_in_place(func, v, *args, **kwargs)
            else:
                _nested[k] = v
    elif isinstance(_nested, list):
        for idx, _l in enumerate(_nested):
            if isinstance(_l, (dict, list, tuple)):
                _nested[idx] = map_nested_keys_in_place(func, _l, *args, **kwargs)
    else:
        raise ValueError(_nested, "Not a nested type")
    return _nested


def is_interactive() -> bool:
    """
    Determine if the current environment is interactive.

    Returns:
        bool: True if interactive (e.g., REPL, IPython, Jupyter Notebook), False if running as a script.
    """
    import __main__ as main

    # When running as a script, the __main__ module has a __file__ attribute.
    # In an interactive environment, the __file__ attribute is absent.
    return not hasattr(main, "__file__")


def dict_remove_nones_in_place(d: Dict[Any, Any]) -> Dict[Any, Any]:
    for k in list(d.keys()):
        if d[k] is None:
            del d[k]
    return d


def simple_repr(object_name: str, **kwargs: Any) -> str:
    """Create a simple string representation of an object.

    For example:

        s = simple_repr("Resource", name="my_resource", table_name="my_table")
        print(s)
        # "Resource(name='my_resource', table_name='my_table')"

    """
    args = [f"{k}={v.__repr__()}" for k, v in kwargs.items()]
    return f"<{object_name}({', '.join(args)})>"


@contextmanager
def custom_environ(env: StrStr) -> Iterator[None]:
    """Temporarily set environment variables inside the context manager and
    fully restore previous environment afterwards
    """
    original_env = {key: os.getenv(key) for key in env}
    os.environ.update(env)
    try:
        yield
    finally:
        for key, value in original_env.items():
            if value is None:
                del os.environ[key]
            else:
                os.environ[key] = value


def with_custom_environ(f: TFun) -> TFun:
    @wraps(f)
    def _wrap(*args: Any, **kwargs: Any) -> Any:
        saved_environ = os.environ.copy()
        try:
            return f(*args, **kwargs)
        finally:
            os.environ.clear()
            os.environ.update(saved_environ)

    return _wrap  # type: ignore


def encoding_for_mode(mode: str) -> Optional[str]:
    if "b" in mode:
        return None
    else:
        return "utf-8"


def main_module_file_path() -> str:
    if len(sys.argv) > 0 and os.path.isfile(sys.argv[0]):
        return str(Path(sys.argv[0]))
    return None


@contextmanager
def set_working_dir(path: str) -> Iterator[str]:
    curr_dir = os.path.abspath(os.getcwd())
    try:
        if path:
            os.chdir(path)
        yield path
    finally:
        os.chdir(curr_dir)


@contextmanager
def multi_context_manager(managers: Sequence[ContextManager[Any]]) -> Iterator[Any]:
    """A context manager holding several other context managers. Enters and exists all of them. Yields from the last in the list"""
    try:
        rv: Any = None
        for manager in managers:
            rv = manager.__enter__()
        yield rv
    except Exception as ex:
        # release context manager
        for manager in managers:
            if isinstance(ex, StopIteration):
                manager.__exit__(None, None, None)
            else:
                manager.__exit__(type(ex), ex, None)
        raise
    else:
        for manager in managers:
            manager.__exit__(None, None, None)


def get_callable_name(f: AnyFun, name_attr: str = "__name__") -> Optional[str]:
    # check first if __name__ is present directly (function), if not then look for type name
    name: str = getattr(f, name_attr, None)
    if not name:
        name = getattr(f.__class__, name_attr, None)
    return name


def is_inner_callable(f: AnyFun) -> bool:
    """Checks if f is defined within other function"""
    # inner functions have full nesting path in their qualname
    return "<locals>" in get_callable_name(f, name_attr="__qualname__")


def get_full_obj_class_name(obj: Any) -> str:
    cls = obj.__class__
    module = cls.__module__
    # exclude 'builtins' for built-in types.
    if module is None or module == "builtins":
        return cls.__name__  #  type: ignore[no-any-return]
    return module + "." + cls.__name__  #  type: ignore[no-any-return]


def get_full_callable_name(f: AnyFun) -> str:
    module = f.__module__
    name = get_callable_name(f, name_attr="__qualname__")
    # exclude 'builtins' for built-in types.
    if module is None or module == "builtins":
        return name
    return module + "." + name


def obfuscate_pseudo_secret(pseudo_secret: str, pseudo_key: bytes) -> str:
    return base64.b64encode(
        bytes([_a ^ _b for _a, _b in zip(pseudo_secret.encode("utf-8"), pseudo_key * 250)])
    ).decode()


def reveal_pseudo_secret(obfuscated_secret: str, pseudo_key: bytes) -> str:
    return bytes(
        [
            _a ^ _b
            for _a, _b in zip(
                base64.b64decode(obfuscated_secret.encode("ascii"), validate=True), pseudo_key * 250
            )
        ]
    ).decode("utf-8")


def get_module_name(m: ModuleType) -> str:
    """Gets module name from module with a fallback for executing module __main__"""
    if m.__name__ == "__main__" and hasattr(m, "__file__"):
        module_file = os.path.basename(m.__file__)
        module_name, _ = os.path.splitext(module_file)
        return module_name
    return m.__name__.split(".")[-1]


def derives_from_class_of_name(o: object, name: str) -> bool:
    """Checks if object o has class of name in its derivation tree"""
    mro = type.mro(type(o))
    return any(t.__name__ == name for t in mro)


def compressed_b64encode(value: bytes) -> str:
    """Compress and b64 encode the given bytestring"""
    return base64.b64encode(zlib.compress(value, level=9)).decode("ascii")


def compressed_b64decode(value: str) -> bytes:
    """Decode a bytestring encoded with `compressed_b64encode`"""
    value_bytes = base64.b64decode(value, validate=True)
    return zlib.decompress(value_bytes)


def identity(x: TAny) -> TAny:
    return x


def increase_row_count(row_counts: RowCounts, counter_name: str, count: int) -> None:
    row_counts[counter_name] = row_counts.get(counter_name, 0) + count


def merge_row_counts(row_counts_1: RowCounts, row_counts_2: RowCounts) -> None:
    """merges row counts_2 into row_counts_1"""
    # only keys present in row_counts_2 are modifed
    for counter_name in row_counts_2.keys():
        row_counts_1[counter_name] = row_counts_1.get(counter_name, 0) + row_counts_2[counter_name]


def extend_list_deduplicated(
    original_list: List[Any],
    extending_list: Iterable[Any],
    normalize_f: Callable[[str], str] = str.__call__,
) -> List[Any]:
    """extends the first list by the second, but does not add duplicates"""
    list_keys = set(normalize_f(s) for s in original_list)
    for item in extending_list:
        if normalize_f(item) not in list_keys:
            original_list.append(item)
    return original_list


@contextmanager
def maybe_context(manager: ContextManager[TAny]) -> Iterator[TAny]:
    """Allows context manager `manager` to be None by creating dummy context. Otherwise `manager` is used"""
    if manager is None:
        yield None
    else:
        with manager as ctx:
            yield ctx


def without_none(d: Mapping[TKey, Optional[TValue]]) -> Mapping[TKey, TValue]:
    """Return a new dict with all `None` values removed"""
    return {k: v for k, v in d.items() if v is not None}


def exclude_keys(mapping: Mapping[str, Any], keys: Iterable[str]) -> Dict[str, Any]:
    """Create a new dictionary from the input mapping, excluding specified keys.

    Args:
        mapping (Mapping[str, Any]): The input mapping from which keys will be excluded.
        keys (Iterable[str]): The keys to exclude.

    Returns:
        Dict[str, Any]: A new dictionary containing all key-value pairs from the original
                        mapping except those with keys specified in `keys`.
    """
    return {k: v for k, v in mapping.items() if k not in keys}


def get_exception_trace(exc: BaseException) -> ExceptionTrace:
    """Get exception trace and additional information for DltException(s)"""
    trace: ExceptionTrace = {"message": str(exc), "exception_type": get_full_obj_class_name(exc)}
    if exc.__traceback__:
        tb_extract = traceback.extract_tb(exc.__traceback__)
        trace["stack_trace"] = traceback.format_list(tb_extract)
    trace["is_terminal"] = isinstance(exc, TerminalException)

    # get attrs and other props
    if isinstance(exc, DltException):
        if exc.__doc__:
            trace["docstring"] = exc.__doc__
        attrs = exc.attrs()
        str_attrs = {}
        for k, v in attrs.items():
            if v is None:
                continue
            try:
                from dlt.common.json import json

                # must be json serializable, other attrs are skipped
                # TODO: be more picky re. what is serialized. exceptions contain data items and
                #  other value that should not be serialized
                if not isinstance(v, str):
                    json.dumps(v)
                str_attrs[k] = v
            except Exception:
                continue
            # extract special attrs
            if k in ["load_id", "pipeline_name", "source_name", "resource_name", "job_id"]:
                trace[k] = v  # type: ignore[literal-required]

        trace["exception_attrs"] = str_attrs
    return trace


def get_exception_trace_chain(
    exc: BaseException, traces: List[ExceptionTrace] = None, seen: Set[int] = None
) -> List[ExceptionTrace]:
    """Get traces for exception chain. The function will recursively visit all __cause__ and __context__ exceptions. The top level
    exception trace is first on the list
    """
    traces = traces or []
    seen = seen or set()
    # prevent cycles
    if id(exc) in seen:
        return traces
    seen.add(id(exc))
    traces.append(get_exception_trace(exc))
    if exc.__cause__:
        return get_exception_trace_chain(exc.__cause__, traces, seen)
    elif exc.__context__:
        return get_exception_trace_chain(exc.__context__, traces, seen)
    return traces


def group_dict_of_lists(input_dict: Dict[str, List[TAny]]) -> List[Dict[str, TAny]]:
    """Decomposes a dictionary with list values into a list of dictionaries with unique keys.

    This function takes an input dictionary where each key maps to a list of objects.
    It returns a list of dictionaries, each containing at most one object per key.
    The goal is to ensure that no two objects with the same key appear in the same dictionary.

    Parameters:
        input_dict (Dict[str, List[Any]]): A dictionary with string keys and list of objects as values.

    Returns:
        List[Dict[str, Any]]: A list of dictionaries, each with unique keys and single objects.
    """
    max_length = max(len(v) for v in input_dict.values())
    list_of_dicts: List[Dict[str, Any]] = [{} for _ in range(max_length)]
    for name, value_list in input_dict.items():
        for idx, obj in enumerate(value_list):
            list_of_dicts[idx][name] = obj
    return list_of_dicts


def order_deduped(lst: List[Any]) -> List[Any]:
    """Returns deduplicated list preserving order of input elements.

    Only works for lists with hashable elements.
    """
    return list(dict.fromkeys(lst))


def assert_min_pkg_version(pkg_name: str, version: str, msg: str = "") -> None:
    version_found = pkg_version(pkg_name)
    if Version(version_found) < Version(version):
        raise DependencyVersionException(
            pkg_name=pkg_name,
            version_found=version_found,
            version_required=">=" + version,
            appendix=msg,
        )


def make_defunct_class(cls: TObj) -> Type[TObj]:
    class DefunctClass(cls.__class__):  # type: ignore[name-defined]
        """A defunct class to replace __class__ when we want to destroy current instance"""

        def __getattribute__(self, name: str) -> Any:
            if name == "__class__":
                # Allow access to __class__
                return object.__getattribute__(self, name)
            else:
                raise RuntimeError("This instance has been dropped and cannot be used anymore.")

    return DefunctClass


class classlocal(Generic[TAny]):
    """A descriptor that is *only* visible on the class that defines it.

    Any attempt to read it from a derived class raises AttributeError,
    so `hasattr(Sub, name)` is *False* unless Sub defines its own copy.
    """

    def __init__(self, value: TAny, owner: type[Any]) -> None:
        self._value: TAny = value
        self._owner: type[Any] = owner

    def __get__(self, obj: Any, cls: type[Any] = None) -> TAny:
        # `cls` is the *class* through which the attribute lookup happened
        # access from the defining class
        if cls is self._owner:
            return self._value
        # access from a subclass or from the descriptor itself: hide it
        raise AttributeError("<class-local>")


def is_typeerror_due_to_wrong_call(exc: Exception, func: AnyFun) -> bool:
    """
    Determine if a TypeError is due to a wrong call to the function (incorrect arguments)
    by inspecting the exception message.
    """
    if not isinstance(exc, TypeError):
        return False
    func_name = func.__name__
    message = str(exc)
    return message.__contains__(f"{func_name}()")


removeprefix = getattr(
    str, "removeprefix", lambda s_, p_: s_[len(p_) :] if s_.startswith(p_) else s_
)


def read_dialect_and_sql(
    file_obj: IO[str],
    fallback_dialect: Optional[TSqlGlotDialect] = None,
) -> Tuple[TSqlGlotDialect, str]:
    """
    Read the first line of a file for the dialect (after the first colon),
    falls back to `fallback_dialect` if not found or empty,
    and then reads the rest as the SQL statement.

    Args:
        file_obj (IO[str]): A file-like object opened in text mode.
        fallback_dialect (Optional[str]): A fallback dialect to use if the first line
            does not specify a dialect.

    Returns:
        Tuple[str, str]: A tuple containing:
            - The extracted or fallback dialect as a string.
            - The SQL statement read from the rest of the file.
    """
    first_line = file_obj.readline()
    # e.g. something like: "dialect: clickhouse\n"
    parts = first_line.split(":", 1)
    parsed_dialect = cast(TSqlGlotDialect, parts[1].strip() if len(parts) > 1 else "")

    dialect = parsed_dialect if parsed_dialect else fallback_dialect

    sql_statement = file_obj.read()
    return dialect, sql_statement



from typing import Optional

from dlt.common.destination.capabilities import DestinationCapabilitiesContext
from dlt.destinations.impl.duckdb.sql_client import DuckDbSqlClient
from dlt.destinations.impl.motherduck.configuration import MotherDuckCredentials


class MotherDuckSqlClient(DuckDbSqlClient):
    def __init__(
        self,
        dataset_name: str,
        staging_dataset_name: str,
        credentials: MotherDuckCredentials,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        super().__init__(dataset_name, staging_dataset_name, credentials, capabilities)
        self.database_name = credentials.database

    def catalog_name(self, quote: bool = True, casefold: bool = True) -> Optional[str]:
        if casefold:
            database_name = self.capabilities.casefold_identifier(self.database_name)
        else:
            database_name = self.database_name
        if quote:
            database_name = self.capabilities.escape_identifier(database_name)
        return database_name



from dlt.common.normalizers.naming.sql_cs_v1 import NamingConvention as SqlCsNamingConvention


class NamingConvention(SqlCsNamingConvention):
    """A variant of sql_cs which lower cases all identifiers."""

    def normalize_identifier(self, identifier: str) -> str:
        return super().normalize_identifier(identifier).lower()

    @property
    def is_case_sensitive(self) -> bool:
        return False



from typing import Any, Optional, Type, Union, Dict, TYPE_CHECKING, Sequence, cast

from dlt.common import logger
from dlt.common.destination.capabilities import TLoaderParallelismStrategy
from dlt.common.destination.exceptions import DestinationException
from dlt.common.exceptions import TerminalValueError
from dlt.common.normalizers.naming.naming import NamingConvention
from dlt.common.reflection.exceptions import ReferenceImportError
from dlt.common.reflection.ref import ImportTrace, callable_typechecker, object_from_ref
from dlt.common.typing import AnyFun
from dlt.common.destination import Destination, DestinationCapabilitiesContext, TLoaderFileFormat
from dlt.common.configuration import known_sections, with_config, get_fun_spec
from dlt.common.configuration.exceptions import ConfigurationValueError
from dlt.common.utils import get_callable_name, is_inner_callable

from dlt.destinations.impl.destination.configuration import (
    CustomDestinationClientConfiguration,
    dummy_custom_destination,
    TDestinationCallable,
)

if TYPE_CHECKING:
    from dlt.destinations.impl.destination.destination import DestinationClient


class UnknownCustomDestinationCallable(ReferenceImportError, DestinationException, KeyError):
    def __init__(
        self, ref: str, qualified_refs: Sequence[str], traces: Sequence[ImportTrace]
    ) -> None:
        self.ref = ref
        self.qualified_refs = qualified_refs
        super().__init__(traces=traces)

    def __str__(self) -> str:
        if "." in self.ref:
            msg = f"Custom destination callable {self.ref} could not be imported."

        if len(self.qualified_refs) == 1 and self.qualified_refs[0] == self.ref:
            pass
        else:
            msg += (
                " Following fully qualified refs were tried in the registry:\n\t%s\n"
                % "\n\t".join(self.qualified_refs)
            )
        if self.traces:
            msg += super().__str__()
        return msg


class destination(Destination[CustomDestinationClientConfiguration, "DestinationClient"]):
    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        caps = DestinationCapabilitiesContext.generic_capabilities("typed-jsonl")
        caps.supported_loader_file_formats = ["typed-jsonl", "parquet"]
        caps.supports_ddl_transactions = False
        caps.supports_transactions = False
        caps.naming_convention = "direct"
        caps.max_table_nesting = 0
        caps.max_parallel_load_jobs = 0
        caps.loader_parallelism_strategy = None
        return caps

    @property
    def spec(self) -> Type[CustomDestinationClientConfiguration]:
        """A spec of destination configuration resolved from the sink function signature"""
        return self._spec

    @property
    def client_class(self) -> Type["DestinationClient"]:
        from dlt.destinations.impl.destination.destination import DestinationClient

        return DestinationClient

    def __init__(
        self,
        destination_callable: Union[TDestinationCallable, str] = None,  # noqa: A003
        destination_name: str = None,
        environment: str = None,
        loader_file_format: TLoaderFileFormat = None,
        batch_size: int = 10,
        naming_convention: str = "direct",
        spec: Type[CustomDestinationClientConfiguration] = None,
        **kwargs: Any,
    ) -> None:
        """Configure the destination to use in a pipeline.

        The dlt.destination decorator wraps this and should preferably be used.

        All arguments provided here supersede other configuration sources such as environment variables and dlt config files.

        Args:
            destination_callable (Union[TDestinationCallable, str], optional): The destination callable to use.
            destination_name (str, optional): Name of the destination, can be used in config section to differentiate between multiple of the same type
            environment (str, optional): Environment of the destination
            loader_file_format (TLoaderFileFormat, optional): The file format to use for the loader
            batch_size (int, optional): The batch size to use for the loader
            naming_convention (str, optional): The naming convention to use for the loader
            spec (Type[CustomDestinationClientConfiguration], optional): The spec to use for the destination
            **kwargs (Any): Additional arguments passed to the destination config
        """
        if spec and not issubclass(spec, CustomDestinationClientConfiguration):
            raise TerminalValueError(
                "A SPEC for a sink destination must use `CustomDestinationClientConfiguration` as a"
                " base."
            )
        # resolve callable
        if callable(destination_callable):
            pass
        elif destination_callable:
            imported_callable, trace = object_from_ref(destination_callable, callable_typechecker)
            if imported_callable is None:
                raise UnknownCustomDestinationCallable(
                    destination_callable, [destination_callable], [trace]
                )
            else:
                destination_callable = imported_callable
        # provide dummy callable for cases where no callable is provided
        # this is needed for cli commands to work
        if not destination_callable:
            logger.warning(
                "No destination callable provided, providing dummy callable which will fail on"
                " load."
            )
            destination_callable = dummy_custom_destination
        elif not callable(destination_callable):
            raise ConfigurationValueError("Resolved Sink destination callable is not a callable.")
        if not destination_name:
            destination_name = destination_callable.__name__

        # build destination spec
        destination_sections = (known_sections.DESTINATION, destination_name)
        conf_callable = with_config(
            destination_callable,
            spec=spec,
            sections=destination_sections,
            include_defaults=True,
            base=None if spec else CustomDestinationClientConfiguration,
        )

        # save destination in registry
        resolved_spec = cast(
            Type[CustomDestinationClientConfiguration], get_fun_spec(conf_callable)
        )
        # remember spec
        self._spec = resolved_spec or spec
        super().__init__(
            destination_name=destination_name,
            environment=environment,
            # NOTE: `loader_file_format` is not a field in the caps so we had to hack the base class to allow this
            loader_file_format=loader_file_format,
            batch_size=batch_size,
            naming_convention=naming_convention,
            destination_callable=conf_callable,
            **kwargs,
        )

    @classmethod
    def adjust_capabilities(
        cls,
        caps: DestinationCapabilitiesContext,
        config: CustomDestinationClientConfiguration,
        naming: Optional[NamingConvention],
    ) -> DestinationCapabilitiesContext:
        caps = super().adjust_capabilities(caps, config, naming)
        caps.preferred_loader_file_format = config.loader_file_format
        return caps


destination.register()



import asyncio
from concurrent.futures import (
    ThreadPoolExecutor,
    as_completed,
    wait as wait_for_futures,
)
import contextvars
from threading import Thread
from typing import Awaitable, Dict, Optional

from dlt.common import logger
from dlt.common.exceptions import PipelineException
from dlt.common.configuration.container import Container
from dlt.common.runners.pool_runner import TimeoutThreadPoolExecutor
from dlt.common.runtime.signals import sleep
from dlt.extract.items import DataItemWithMeta, TItemFuture, ResolvablePipeItem, FuturePipeItem

from dlt.extract.exceptions import (
    DltSourceException,
    ExtractorException,
    PipeException,
    ResourceExtractionError,
)


class FuturesPool:
    """Worker pool for pipe items that can be resolved asynchronously.

    Items can be either asyncio coroutines or regular callables which will be executed in a thread pool.
    """

    def __init__(
        self, workers: int = 5, poll_interval: float = 0.01, max_parallel_items: int = 20
    ) -> None:
        self.futures: Dict[TItemFuture, FuturePipeItem] = {}
        self._thread_pool: ThreadPoolExecutor = None
        self._async_pool: asyncio.AbstractEventLoop = None
        self._async_pool_thread: Thread = None
        self.workers = workers
        self.poll_interval = poll_interval
        self.max_parallel_items = max_parallel_items
        self.used_slots: int = 0

    def __len__(self) -> int:
        return len(self.futures)

    @property
    def free_slots(self) -> int:
        # Done futures don't count as slots, so we can still add futures
        return self.max_parallel_items - self.used_slots

    @property
    def empty(self) -> bool:
        return len(self.futures) == 0

    def _ensure_thread_pool(self) -> ThreadPoolExecutor:
        # lazily start or return thread pool
        if self._thread_pool:
            return self._thread_pool

        self._thread_pool = TimeoutThreadPoolExecutor(
            self.workers,
            timeout=2.0,
            thread_name_prefix=Container.thread_pool_prefix() + "extract-threads",
        )
        return self._thread_pool

    def _ensure_async_pool(self) -> asyncio.AbstractEventLoop:
        # lazily create async pool is separate thread
        if self._async_pool:
            return self._async_pool

        def start_background_loop(loop: asyncio.AbstractEventLoop) -> None:
            asyncio.set_event_loop(loop)
            loop.run_forever()

        self._async_pool = asyncio.new_event_loop()
        self._async_pool_thread = Thread(
            target=start_background_loop,
            args=(self._async_pool,),
            daemon=True,
            name=Container.thread_pool_prefix() + "extract-futures",
        )
        self._async_pool_thread.start()

        # start or return async pool
        return self._async_pool

    def _vacate_slot(self, _: TItemFuture) -> None:
        # Used as callback to free up slot when future is done
        self.used_slots -= 1

    def submit(self, pipe_item: ResolvablePipeItem) -> TItemFuture:
        """Submit an item to the pool.

        Args:
            pipe_item: The pipe item to submit. `pipe_item.item` must be either an asyncio coroutine or a callable.

        Returns:
            The resulting future object
        """

        # Sanity check, negative free slots means there's a bug somewhere
        assert self.free_slots >= 0, "Worker pool has negative free slots, this should never happen"

        if self.free_slots == 0:
            # Wait until some future is completed to ensure there's a free slot
            # Note: This is probably not thread safe. If ever multiple threads will be submitting
            # jobs to the pool, we ned to change this whole method to be inside a `threading.Lock`
            self._wait_for_free_slot()

        future: Optional[TItemFuture] = None

        # submit to thread pool or async pool
        item = pipe_item.item
        if isinstance(item, Awaitable):
            future = asyncio.run_coroutine_threadsafe(item, self._ensure_async_pool())
        elif callable(item):
            # pass pipe context to thread pool, happens automatically for coroutines
            ctx = contextvars.copy_context()
            future = self._ensure_thread_pool().submit(ctx.run, item)
        else:
            raise ValueError(f"Unsupported item type: `{type(item)}`")

        # Future is not removed from self.futures until it's been consumed by the
        # pipe iterator. But we always want to vacate a slot so new jobs can be submitted
        future.add_done_callback(self._vacate_slot)
        self.used_slots += 1

        self.futures[future] = FuturePipeItem(
            future, pipe_item.step, pipe_item.pipe, pipe_item.meta
        )
        return future

    def sleep(self) -> None:
        sleep(self.poll_interval)

    def _resolve_future(self, future: TItemFuture) -> Optional[ResolvablePipeItem]:
        future, step, pipe, meta = self.futures.pop(future)

        if ex := future.exception():
            if isinstance(ex, StopAsyncIteration):
                return None
            # Raise if any future fails
            if isinstance(
                ex, (PipelineException, ExtractorException, DltSourceException, PipeException)
            ):
                raise ex
            raise ResourceExtractionError(pipe.name, future, str(ex), "future") from ex

        item = future.result()

        if item is None:
            return None
        elif isinstance(item, DataItemWithMeta):
            return ResolvablePipeItem(item.data, step, pipe, item.meta)
        else:
            return ResolvablePipeItem(item, step, pipe, meta)

    def _next_done_future(self) -> Optional[TItemFuture]:
        """Get the done future in the pool (if any). This does not block."""
        return next((fut for fut in self.futures if fut.done() and not fut.cancelled()), None)

    def resolve_next_future(
        self, use_configured_timeout: bool = False
    ) -> Optional[ResolvablePipeItem]:
        """Block until the next future is done and return the result. Returns None if no futures done.

        Args:
            use_configured_timeout: If True, use the value of `self.poll_interval` as the max wait time,
                raises `concurrent.futures.TimeoutError` if no future is done within that time.

        Returns:
            The resolved future item or None if no future is done.
        """
        if not self.futures:
            return None

        if (future := self._next_done_future()) is not None:
            # When there are multiple already done futures from the same pipe we return results in insertion order
            return self._resolve_future(future)
        for future in as_completed(
            self.futures, timeout=self.poll_interval if use_configured_timeout else None
        ):
            if future.cancelled():
                # Get the next not-cancelled future
                continue

            return self._resolve_future(future)

        return None

    def resolve_next_future_no_wait(self) -> Optional[ResolvablePipeItem]:
        """Resolve the first done future in the pool.
        This does not block and returns None if no future is done.
        """
        # Get next done future
        future = self._next_done_future()
        if not future:
            return None

        return self._resolve_future(future)

    def _wait_for_free_slot(self) -> None:
        """Wait until any future in the pool is completed to ensure there's a free slot."""
        if self.free_slots >= 1:
            return

        for future in as_completed(self.futures):
            if future.cancelled():
                # Get the next not-cancelled future
                continue
            if self.free_slots == 0:
                # Future was already completed so slot was not freed
                continue
            return  # Return when first future completes

    def close(self) -> None:
        # Cancel all futures
        for f in self.futures:
            if not f.done():
                f.cancel()

        def stop_background_loop(loop: asyncio.AbstractEventLoop) -> None:
            loop.stop()

        if self._async_pool:
            # wait for all async generators to be closed
            future = asyncio.run_coroutine_threadsafe(
                self._async_pool.shutdown_asyncgens(), self._ensure_async_pool()
            )

            _, not_done = wait_for_futures([future], timeout=60)
            if not_done:
                logger.warning(
                    "Closing extract async pool: shutdown_asyncgens did not complete"
                    " within 60s, abandoning remaining async generators"
                )
            self._async_pool.call_soon_threadsafe(stop_background_loop, self._async_pool)

            self._async_pool_thread.join(timeout=10)
            if self._async_pool_thread.is_alive():
                logger.warning(
                    "Closing extract async pool: event loop thread did not stop"
                    " within 10s, abandoning"
                )
            else:
                self._async_pool.close()
            self._async_pool = None
            self._async_pool_thread = None

        if self._thread_pool:
            self._thread_pool.shutdown(wait=True)
            self._thread_pool = None

        self.futures.clear()



import contextlib
import os
from copy import deepcopy
import threading

import datetime  # noqa: 251
import humanize
from pathlib import PurePath
from pendulum.datetime import DateTime
from typing import (
    ClassVar,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Literal,
    Optional,
    Sequence,
    Set,
    cast,
    Any,
    Tuple,
)
from typing_extensions import NotRequired

from dlt.common.typing import TypedDict, get_args, DictStrAny, SupportsHumanize
from dlt.common.pendulum import pendulum
from dlt.common.json import json
from dlt.common.configuration import configspec
from dlt.common.configuration.specs import ContainerInjectableContext
from dlt.common.configuration.exceptions import ContextDefaultCannotBeCreated
from dlt.common.configuration.container import Container
from dlt.common.data_writers import DataWriter, new_file_id
from dlt.common.destination import TLoaderFileFormat
from dlt.common.exceptions import TerminalValueError
from dlt.common.schema import Schema, TSchemaTables
from dlt.common.schema.typing import TStoredSchema, TTableSchemaColumns, TTableSchema
from dlt.common.storages import FileStorage
from dlt.common.storages.exceptions import (
    LoadPackageAlreadyCompleted,
    LoadPackageCancelled,
    LoadPackageNotCompleted,
    LoadPackageNotFound,
    CurrentLoadPackageStateNotAvailable,
)
from dlt.common.utils import flatten_list_or_items
from dlt.common.versioned_state import (
    generate_state_version_hash,
    bump_state_version_if_modified,
    TVersionedState,
    default_versioned_state,
    json_decode_state,
    json_encode_state,
)
from dlt.common.time import precise_time, increasing_precise_time

TJobFileFormat = Literal["sql", "reference", TLoaderFileFormat]
"""Loader file formats with internal job types"""
JOB_EXCEPTION_EXTENSION = ".exception"


class TPipelineStateDoc(TypedDict, total=False):
    """Corresponds to the StateInfo Tuple"""

    version: int
    engine_version: int
    pipeline_name: str
    state: str
    created_at: datetime.datetime
    version_hash: str
    _dlt_load_id: NotRequired[str]


class TLoadPackageDropTablesState(TypedDict):
    dropped_tables: NotRequired[List[TTableSchema]]
    """List of tables that are to be dropped from the schema and destination (i.e. when `refresh` mode is used)"""
    truncated_tables: NotRequired[List[TTableSchema]]
    """List of tables that are to be truncated in the destination (i.e. when `refresh='drop_data'` mode is used)"""


class TLoadPackageState(TVersionedState, TLoadPackageDropTablesState, total=False):
    created_at: DateTime
    """Timestamp when the load package was created"""
    pipeline_state: NotRequired[TPipelineStateDoc]
    """Pipeline state, added at the end of the extraction phase"""

    """A section of state that does not participate in change merging and version control"""
    destination_state: NotRequired[Dict[str, Any]]
    """private space for destinations to store state relevant only to the load package"""


class TLoadPackage(TypedDict, total=False):
    load_id: str
    """Load id"""
    state: TLoadPackageState
    """State of the load package"""


# allows to upgrade state when restored with a new version of state logic/schema
LOAD_PACKAGE_STATE_ENGINE_VERSION = 1


def generate_loadpackage_state_version_hash(state: TLoadPackageState) -> str:
    return generate_state_version_hash(state)


def bump_loadpackage_state_version_if_modified(state: TLoadPackageState) -> Tuple[int, str, str]:
    return bump_state_version_if_modified(state)


def migrate_load_package_state(
    state: DictStrAny, from_engine: int, to_engine: int
) -> TLoadPackageState:
    # TODO: if you start adding new versions, we need proper tests for these migrations!
    # NOTE: do not touch destinations state, it is not versioned
    if from_engine == to_engine:
        return cast(TLoadPackageState, state)

    # check state engine
    if from_engine != to_engine:
        raise Exception("No upgrade path for loadpackage state")

    state["_state_engine_version"] = from_engine
    return cast(TLoadPackageState, state)


def default_load_package_state() -> TLoadPackageState:
    return {
        **default_versioned_state(),
        "_state_engine_version": LOAD_PACKAGE_STATE_ENGINE_VERSION,
    }


def create_load_id() -> str:
    """Creates new package load id which is the current unix timestamp converted to string.
    Load ids must have the following properties:
    - They must maintain increase order over time for a particular dlt schema loaded to particular destination and dataset
    `dlt` executes packages in order of load ids
    `dlt` considers a state with the highest load id to be the most up to date when restoring state from destination
    """
    return str(increasing_precise_time())


# folders to manage load jobs in a single load package
TPackageJobState = Literal["new_jobs", "failed_jobs", "started_jobs", "completed_jobs"]
WORKING_FOLDERS: Set[TPackageJobState] = set(get_args(TPackageJobState))
TLoadPackageStatus = Literal["new", "extracted", "normalized", "loaded", "aborted"]


class ParsedLoadJobFileName(NamedTuple):
    """Represents a file name of a job in load package. The file name contains name of a table, number of times the job was retried, extension
    and a 5 bytes random string to make job file name unique.
    The job id does not contain retry count and is immutable during loading of the data
    """

    table_name: str
    file_id: str
    retry_count: int
    file_format: TJobFileFormat
    is_compressed: bool = False

    def job_id(self) -> str:
        """Unique identifier of the job"""
        compression_extension = ".gz" if self.is_compressed else ""
        return f"{self.table_name}.{self.file_id}.{self.file_format}{compression_extension}"

    def file_name(self) -> str:
        """A name of the file with the data to be loaded"""
        compression_extension = ".gz" if self.is_compressed else ""
        return f"{self.table_name}.{self.file_id}.{int(self.retry_count)}.{self.file_format}{compression_extension}"

    def with_retry(self) -> "ParsedLoadJobFileName":
        """Returns a job with increased retry count"""
        return self._replace(retry_count=self.retry_count + 1)

    def to_reference_file_name(self) -> str:
        """Returns a file name for a reference job"""
        return f"{self.table_name}.{self.file_id}.{self.retry_count}.reference"

    def full_extension(self) -> str:
        """Returns the full file extension"""
        return f"{self.file_format}.gz" if self.is_compressed else f"{self.file_format}"

    @staticmethod
    def parse(file_name: str) -> "ParsedLoadJobFileName":
        p = PurePath(file_name)
        parts = p.name.split(".")

        if len(parts) == 4:
            # No compression extension: table_name.file_id.retry_count.file_format
            return ParsedLoadJobFileName(
                parts[0], parts[1], int(parts[2]), cast(TJobFileFormat, parts[3]), False
            )
        elif len(parts) == 5 and parts[4] == "gz":
            # With compression extension: table_name.file_id.retry_count.file_format.gz
            return ParsedLoadJobFileName(
                parts[0], parts[1], int(parts[2]), cast(TJobFileFormat, parts[3]), True
            )
        else:
            raise TerminalValueError(parts)

    @staticmethod
    def new_file_id() -> str:
        return new_file_id()

    def __str__(self) -> str:
        return self.job_id()


class LoadJobInfo(NamedTuple):
    state: TPackageJobState
    file_path: str
    file_size: int
    created_at: datetime.datetime
    elapsed: float
    job_file_info: ParsedLoadJobFileName
    failed_message: str

    def asdict(self) -> DictStrAny:
        d = self._asdict()
        # flatten
        del d["job_file_info"]
        d.update(self.job_file_info._asdict())
        d["job_id"] = self.job_file_info.job_id()
        return d

    def asstr(self, verbosity: int = 0) -> str:
        failed_msg = (
            "The job FAILED TERMINALLY and cannot be restarted." if self.failed_message else ""
        )
        elapsed_msg = (
            humanize.precisedelta(pendulum.duration(seconds=self.elapsed))
            if self.elapsed
            else "---"
        )
        msg = (
            f"Job: {self.job_file_info.job_id()}, table: {self.job_file_info.table_name} in"
            f" {self.state}. "
        )
        msg += (
            f"File type: {self.job_file_info.file_format}, size:"
            f" {humanize.naturalsize(self.file_size, binary=True, gnu=True)}. "
        )
        msg += f"Started on: {self.created_at} and completed in {elapsed_msg}."
        if failed_msg:
            msg += "\nThe job FAILED TERMINALLY and cannot be restarted."
            if verbosity > 0:
                msg += "\n" + self.failed_message
        return msg

    def __str__(self) -> str:
        return self.asstr(verbosity=0)


class _LoadPackageInfo(NamedTuple):
    load_id: str
    package_path: str
    state: TLoadPackageStatus
    schema: Schema
    schema_update: TSchemaTables
    completed_at: datetime.datetime
    jobs: Dict[TPackageJobState, List[LoadJobInfo]]


class LoadPackageInfo(SupportsHumanize, _LoadPackageInfo):
    @property
    def schema_name(self) -> str:
        return self.schema.name

    @property
    def schema_hash(self) -> str:
        return self.schema.version_hash

    def asdict(self) -> DictStrAny:
        d = self._asdict()
        # job as list
        d["jobs"] = [job.asdict() for job in flatten_list_or_items(iter(self.jobs.values()))]  # type: ignore
        d["schema_hash"] = self.schema_hash
        d["schema_name"] = self.schema_name
        # flatten update into list of columns
        tables: List[DictStrAny] = deepcopy(list(self.schema_update.values()))  # type: ignore
        for table in tables:
            table.pop("filters", None)
            columns: List[DictStrAny] = []
            table["schema_name"] = self.schema_name
            table["load_id"] = self.load_id
            for column in table["columns"].values():
                column["table_name"] = table["name"]
                column["schema_name"] = self.schema_name
                column["load_id"] = self.load_id
                columns.append(column)
            table["columns"] = columns
        d.pop("schema_update")
        d.pop("schema")
        d["tables"] = tables

        return d

    def asstr(self, verbosity: int = 0) -> str:
        completed_msg = (
            f"The package was {self.state.upper()} at {self.completed_at}"
            if self.completed_at
            else "The package is NOT YET LOADED to the destination"
        )
        msg = (
            f"The package with load id {self.load_id} for schema {self.schema_name} is in"
            f" {self.state.upper()} state. It updated schema for {len(self.schema_update)} tables."
            f" {completed_msg}.\n"
        )
        msg += "Jobs details:\n"
        msg += "\n".join(job.asstr(verbosity) for job in flatten_list_or_items(iter(self.jobs.values())))  # type: ignore
        return msg

    def __str__(self) -> str:
        return self.asstr(verbosity=0)


class PackageStorage:
    NEW_JOBS_FOLDER: ClassVar[TPackageJobState] = "new_jobs"
    FAILED_JOBS_FOLDER: ClassVar[TPackageJobState] = "failed_jobs"
    STARTED_JOBS_FOLDER: ClassVar[TPackageJobState] = "started_jobs"
    COMPLETED_JOBS_FOLDER: ClassVar[TPackageJobState] = "completed_jobs"

    SCHEMA_FILE_NAME: ClassVar[str] = "schema.json"
    SCHEMA_UPDATES_FILE_NAME = (  # updates to the tables in schema created by normalizer
        "schema_updates.json"
    )
    APPLIED_SCHEMA_UPDATES_FILE_NAME = (
        "applied_" + "schema_updates.json"
    )  # updates applied to the destination
    PACKAGE_COMPLETED_FILE_NAME = (  # completed package marker file, currently only to store data with os.stat
        "package_completed.json"
    )
    LOAD_PACKAGE_STATE_FILE_NAME = (  # internal state of the load package, will not be synced to the destination
        "load_package_state.json"
    )
    CANCEL_PACKAGE_FILE_NAME = "_cancelled"

    def __init__(self, storage: FileStorage, initial_state: TLoadPackageStatus) -> None:
        """Creates storage that manages load packages with root at `storage` and initial package state `initial_state`"""
        self.storage = storage
        self.initial_state = initial_state

    #
    # List jobs
    #

    def get_package_path(self, load_id: str) -> str:
        """Gets path of the package relative to storage root"""
        return load_id

    def get_job_state_folder_path(self, load_id: str, state: TPackageJobState) -> str:
        """Gets path to the jobs in `state` in package `load_id`, relative to the storage root"""
        return os.path.join(self.get_package_path(load_id), state)

    def get_job_file_path(self, load_id: str, state: TPackageJobState, file_name: str) -> str:
        """Get path to job with `file_name` in `state` in package `load_id`, relative to the storage root"""
        return os.path.join(self.get_job_state_folder_path(load_id, state), file_name)

    def list_packages(self) -> Sequence[str]:
        """Lists all load ids in storage, earliest first

        NOTE: Load ids are sorted alphabetically. This class does not store package creation time separately.
        """
        loads = self.storage.list_folder_dirs(".", to_root=False)
        # start from the oldest packages
        return sorted(loads)

    def list_new_jobs(self, load_id: str) -> Sequence[str]:
        new_jobs = self.storage.list_folder_files(
            self.get_job_state_folder_path(load_id, PackageStorage.NEW_JOBS_FOLDER)
        )
        return new_jobs

    def list_started_jobs(self, load_id: str) -> Sequence[str]:
        return self.storage.list_folder_files(
            self.get_job_state_folder_path(load_id, PackageStorage.STARTED_JOBS_FOLDER)
        )

    def list_failed_jobs(self, load_id: str) -> Sequence[str]:
        return [
            file
            for file in self.storage.list_folder_files(
                self.get_job_state_folder_path(load_id, PackageStorage.FAILED_JOBS_FOLDER)
            )
            if not file.endswith(JOB_EXCEPTION_EXTENSION)
        ]

    def list_job_with_states_for_table(
        self, load_id: str, table_name: str
    ) -> Sequence[Tuple[TPackageJobState, ParsedLoadJobFileName]]:
        return self.filter_jobs_for_table(self.list_all_jobs_with_states(load_id), table_name)

    def list_all_jobs_with_states(
        self, load_id: str
    ) -> Sequence[Tuple[TPackageJobState, ParsedLoadJobFileName]]:
        info = self.get_load_package_jobs(load_id)
        state_jobs = []
        for state, jobs in info.items():
            state_jobs.extend([(state, job) for job in jobs])
        return state_jobs

    def list_failed_jobs_infos(self, load_id: str) -> Sequence[LoadJobInfo]:
        """List all failed jobs and associated error messages for a load package with `load_id`"""
        if not self.is_package_completed(load_id):
            raise LoadPackageNotCompleted(load_id)
        failed_jobs: List[LoadJobInfo] = []
        package_path = self.get_package_path(load_id)
        package_created_at = pendulum.from_timestamp(
            os.path.getmtime(
                self.storage.make_full_path(
                    os.path.join(package_path, PackageStorage.PACKAGE_COMPLETED_FILE_NAME)
                )
            )
        )
        for file in self.list_failed_jobs(load_id):
            failed_jobs.append(
                self._read_job_file_info(
                    load_id, "failed_jobs", ParsedLoadJobFileName.parse(file), package_created_at
                )
            )
        return failed_jobs

    def is_package_completed(self, load_id: str) -> bool:
        package_path = self.get_package_path(load_id)
        return self.storage.has_file(
            os.path.join(package_path, PackageStorage.PACKAGE_COMPLETED_FILE_NAME)
        )

    #
    # Move jobs
    #

    def import_job(
        self, load_id: str, job_file_path: str, job_state: TPackageJobState = "new_jobs"
    ) -> None:
        """Adds new job by moving the `job_file_path` into `new_jobs` of package `load_id`"""
        self.storage.atomic_import(
            job_file_path, self.get_job_state_folder_path(load_id, job_state)
        )

    def start_job(self, load_id: str, file_name: str) -> str:
        return self._move_job(
            load_id, PackageStorage.NEW_JOBS_FOLDER, PackageStorage.STARTED_JOBS_FOLDER, file_name
        )

    def fail_job(self, load_id: str, file_name: str, failed_message: Optional[str]) -> str:
        # save the exception to failed jobs
        if failed_message:
            self.storage.save(
                self.get_job_file_path(
                    load_id, PackageStorage.FAILED_JOBS_FOLDER, file_name + JOB_EXCEPTION_EXTENSION
                ),
                failed_message,
            )
        # move to failed jobs
        return self._move_job(
            load_id,
            PackageStorage.STARTED_JOBS_FOLDER,
            PackageStorage.FAILED_JOBS_FOLDER,
            file_name,
        )

    def retry_job(self, load_id: str, file_name: str) -> str:
        # when retrying job we must increase the retry count
        source_fn = ParsedLoadJobFileName.parse(file_name)
        dest_fn = source_fn.with_retry()
        # move it directly to new file name
        return self._move_job(
            load_id,
            PackageStorage.STARTED_JOBS_FOLDER,
            PackageStorage.NEW_JOBS_FOLDER,
            file_name,
            dest_fn.file_name(),
        )

    def complete_job(self, load_id: str, file_name: str) -> str:
        return self._move_job(
            load_id,
            PackageStorage.STARTED_JOBS_FOLDER,
            PackageStorage.COMPLETED_JOBS_FOLDER,
            file_name,
        )

    #
    # Create and drop entities
    #

    def create_package(
        self, load_id: str, initial_state: TLoadPackageState = None, schema: Schema = None
    ) -> None:
        self.storage.create_folder(load_id)
        # create processing directories
        self.storage.create_folder(os.path.join(load_id, PackageStorage.NEW_JOBS_FOLDER))
        self.storage.create_folder(os.path.join(load_id, PackageStorage.COMPLETED_JOBS_FOLDER))
        self.storage.create_folder(os.path.join(load_id, PackageStorage.FAILED_JOBS_FOLDER))
        self.storage.create_folder(os.path.join(load_id, PackageStorage.STARTED_JOBS_FOLDER))
        # use initial state or create a new by loading non existing state
        state = self.get_load_package_state(load_id) if initial_state is None else initial_state
        if not state.get("created_at"):
            # try to parse load_id as unix timestamp
            try:
                created_at = float(load_id)
            except Exception:
                created_at = precise_time()
            state["created_at"] = pendulum.from_timestamp(created_at)
        self.save_load_package_state(load_id, state)
        if schema:
            self.save_schema(load_id, schema)

    def complete_loading_package(self, load_id: str, load_state: TLoadPackageStatus) -> str:
        """Completes loading the package by writing marker file with`package_state. Returns path to the completed package"""
        load_path = self.get_package_path(load_id)
        if self.is_package_completed(load_id):
            raise LoadPackageAlreadyCompleted(load_id)
        # save marker file
        self.storage.save(
            os.path.join(load_path, PackageStorage.PACKAGE_COMPLETED_FILE_NAME), load_state
        )
        # TODO: also modify state
        return load_path

    def remove_completed_jobs(self, load_id: str) -> None:
        """Deletes completed jobs. If package has failed jobs, nothing gets deleted."""
        has_failed_jobs = len(self.list_failed_jobs(load_id)) > 0
        # delete completed jobs
        if not has_failed_jobs:
            self.storage.delete_folder(
                self.get_job_state_folder_path(load_id, PackageStorage.COMPLETED_JOBS_FOLDER),
                recursively=True,
            )

    def delete_package(self, load_id: str, not_exists_ok: bool = False) -> None:
        package_path = self.get_package_path(load_id)
        if not self.storage.has_folder(package_path):
            if not_exists_ok:
                return
            raise LoadPackageNotFound(load_id)
        self.storage.delete_folder(package_path, recursively=True)

    def load_schema(self, load_id: str) -> Schema:
        return Schema.from_dict(self._load_schema(load_id), validate_schema=False)

    def schema_name(self, load_id: str) -> str:
        """Gets schema name associated with the package"""
        schema_dict: TStoredSchema = self._load_schema(load_id)  # type: ignore[assignment]
        return schema_dict["name"]

    def save_schema(self, load_id: str, schema: Schema) -> str:
        # save a schema to a temporary load package
        dump = json.dumps(schema.to_dict())
        return self.storage.save(os.path.join(load_id, PackageStorage.SCHEMA_FILE_NAME), dump)

    def save_schema_updates(self, load_id: str, schema_update: TSchemaTables) -> None:
        with self.storage.open_file(
            os.path.join(load_id, PackageStorage.SCHEMA_UPDATES_FILE_NAME), mode="wb"
        ) as f:
            json.dump(schema_update, f)

    def cancel(self, load_id: str) -> None:
        """Sets cancel flag currently used for inter-process signalling"""
        package_path = self.get_package_path(load_id)
        if not self.storage.has_folder(package_path):
            raise LoadPackageNotFound(load_id)
        self.storage.touch_file(os.path.join(package_path, self.CANCEL_PACKAGE_FILE_NAME))

    def raise_if_cancelled(self, load_id: str) -> None:
        """Raise an exception if package is cancelled"""
        package_path = self.get_package_path(load_id)
        if not self.storage.has_folder(package_path):
            raise LoadPackageNotFound(load_id)
        if self.storage.has_file(os.path.join(package_path, self.CANCEL_PACKAGE_FILE_NAME)):
            raise LoadPackageCancelled(load_id)

    #
    # Load package state
    #

    def get_load_package_state(self, load_id: str) -> TLoadPackageState:
        package_path = self.get_package_path(load_id)
        if not self.storage.has_folder(package_path):
            raise LoadPackageNotFound(load_id)
        try:
            state_dump = self.storage.load(self.get_load_package_state_path(load_id))
            state = json_decode_state(state_dump)
            return migrate_load_package_state(
                state, state["_state_engine_version"], LOAD_PACKAGE_STATE_ENGINE_VERSION
            )
        except FileNotFoundError:
            return default_load_package_state()

    def save_load_package_state(self, load_id: str, state: TLoadPackageState) -> None:
        package_path = self.get_package_path(load_id)
        if not self.storage.has_folder(package_path):
            raise LoadPackageNotFound(load_id)
        bump_loadpackage_state_version_if_modified(state)
        self.storage.save(
            self.get_load_package_state_path(load_id),
            json_encode_state(state),
        )

    def get_load_package_state_path(self, load_id: str) -> str:
        package_path = self.get_package_path(load_id)
        return os.path.join(package_path, PackageStorage.LOAD_PACKAGE_STATE_FILE_NAME)

    #
    # Get package info
    #

    def get_load_package_jobs(
        self, load_id: str
    ) -> Dict[TPackageJobState, List[ParsedLoadJobFileName]]:
        """Gets all jobs in a package and returns them as lists assigned to a particular state."""
        package_path = self.get_package_path(load_id)
        if not self.storage.has_folder(package_path):
            raise LoadPackageNotFound(load_id)
        all_jobs: Dict[TPackageJobState, List[ParsedLoadJobFileName]] = {}
        for state in WORKING_FOLDERS:
            jobs: List[ParsedLoadJobFileName] = []
            with contextlib.suppress(FileNotFoundError):
                # we ignore if load package lacks one of working folders. completed_jobs may be deleted on archiving
                for file in self.storage.list_folder_files(
                    self.get_job_state_folder_path(load_id, state), to_root=False
                ):
                    if not file.endswith(JOB_EXCEPTION_EXTENSION):
                        jobs.append(ParsedLoadJobFileName.parse(file))
            all_jobs[state] = jobs
        return all_jobs

    def get_load_package_info(self, load_id: str) -> LoadPackageInfo:
        """Gets information on normalized/completed package with given load_id, all jobs and their statuses.

        Will reach to the file system to get additional stats, mtime, also collects exceptions for failed jobs.
        NOTE: do not call this function often. it should be used only to generate metrics
        """
        package_path = self.get_package_path(load_id)
        package_jobs = self.get_load_package_jobs(load_id)

        package_created_at: DateTime = None
        package_state = self.initial_state
        applied_update: TSchemaTables = {}

        # check if package completed
        completed_file_path = os.path.join(package_path, PackageStorage.PACKAGE_COMPLETED_FILE_NAME)
        if self.storage.has_file(completed_file_path):
            package_created_at = pendulum.from_timestamp(
                os.path.getmtime(self.storage.make_full_path(completed_file_path))
            )
            package_state = self.storage.load(completed_file_path)

        # check if schema updates applied
        applied_schema_update_file = os.path.join(
            package_path, PackageStorage.APPLIED_SCHEMA_UPDATES_FILE_NAME
        )
        if self.storage.has_file(applied_schema_update_file):
            applied_update = json.loads(self.storage.load(applied_schema_update_file))
        schema = Schema.from_dict(self._load_schema(load_id), validate_schema=False)

        # read jobs with all statuses
        all_job_infos: Dict[TPackageJobState, List[LoadJobInfo]] = {}
        for state, jobs in package_jobs.items():
            all_job_infos[state] = [
                self._read_job_file_info(load_id, state, job, package_created_at) for job in jobs
            ]

        return LoadPackageInfo(
            load_id,
            self.storage.make_full_path(package_path),
            package_state,
            schema,
            applied_update,
            package_created_at,
            all_job_infos,
        )

    def get_job_failed_message(self, load_id: str, job: ParsedLoadJobFileName) -> str:
        """Get exception message of a failed job."""
        rel_path = self.get_job_file_path(load_id, "failed_jobs", job.file_name())
        if not self.storage.has_file(rel_path):
            raise FileNotFoundError(rel_path)
        failed_message: str = None
        with contextlib.suppress(FileNotFoundError):
            failed_message = self.storage.load(rel_path + JOB_EXCEPTION_EXTENSION)
        return failed_message

    def job_to_job_info(
        self, load_id: str, state: TPackageJobState, job: ParsedLoadJobFileName
    ) -> LoadJobInfo:
        """Creates partial job info by converting job object. size, mtime and failed message will not be populated"""
        full_path = os.path.join(
            self.storage.storage_path, self.get_job_file_path(load_id, state, job.file_name())
        )
        return LoadJobInfo(
            state,
            full_path,
            0,
            None,
            0,
            job,
            None,
        )

    def _read_job_file_info(
        self,
        load_id: str,
        state: TPackageJobState,
        job: ParsedLoadJobFileName,
        now: DateTime = None,
    ) -> LoadJobInfo:
        """Creates job info by reading additional props from storage"""
        failed_message = None
        if state == "failed_jobs":
            failed_message = self.get_job_failed_message(load_id, job)
        full_path = os.path.join(
            self.storage.storage_path, self.get_job_file_path(load_id, state, job.file_name())
        )
        st = os.stat(full_path)
        return LoadJobInfo(
            state,
            full_path,
            st.st_size,
            pendulum.from_timestamp(st.st_mtime),
            PackageStorage._job_elapsed_time_seconds(full_path, now.timestamp() if now else None),
            job,
            failed_message,
        )

    #
    # Utils
    #

    def _move_job(
        self,
        load_id: str,
        source_folder: TPackageJobState,
        dest_folder: TPackageJobState,
        file_name: str,
        new_file_name: str = None,
    ) -> str:
        # ensure we move file names, not paths
        assert file_name == FileStorage.get_file_name_from_file_path(file_name)

        dest_path = self.get_job_file_path(load_id, dest_folder, new_file_name or file_name)
        self.storage.atomic_rename(
            self.get_job_file_path(load_id, source_folder, file_name), dest_path
        )
        return self.storage.make_full_path(dest_path)

    def _load_schema(self, load_id: str) -> DictStrAny:
        schema_path = os.path.join(load_id, PackageStorage.SCHEMA_FILE_NAME)
        return json.loads(self.storage.load(schema_path))  # type: ignore[no-any-return]

    @staticmethod
    def build_job_file_name(
        table_name: str,
        file_id: str,
        retry_count: int = 0,
        validate_components: bool = True,
        loader_file_format: TLoaderFileFormat = None,
    ) -> str:
        if validate_components:
            FileStorage.validate_file_name_component(table_name)
        fn = f"{table_name}.{file_id}.{int(retry_count)}"
        if loader_file_format:
            format_spec = DataWriter.writer_spec_from_file_format(loader_file_format, "object")
            return fn + f".{format_spec.file_extension}"
        return fn

    @staticmethod
    def is_package_partially_loaded(package_info: LoadPackageInfo) -> bool:
        """Checks if package is partially loaded - has jobs that are completed and jobs that are not."""
        all_jobs_count = sum(len(package_info.jobs[job_state]) for job_state in WORKING_FOLDERS)
        completed_jobs_count = len(package_info.jobs["completed_jobs"])
        if completed_jobs_count and all_jobs_count - completed_jobs_count > 0:
            return True
        return False

    @staticmethod
    def _job_elapsed_time_seconds(file_path: str, now_ts: float = None) -> float:
        return (now_ts or precise_time()) - os.path.getmtime(file_path)

    @staticmethod
    def filter_jobs_for_table(
        all_jobs: Iterable[Tuple[TPackageJobState, ParsedLoadJobFileName]], table_name: str
    ) -> Sequence[Tuple[TPackageJobState, ParsedLoadJobFileName]]:
        return [job for job in all_jobs if job[1].table_name == table_name]


@configspec
class LoadPackageStateInjectableContext(ContainerInjectableContext):
    storage: PackageStorage = None
    load_id: str = None
    can_create_default: ClassVar[bool] = False
    global_affinity: ClassVar[bool] = False

    def commit(self) -> None:
        with self.state_save_lock:
            self.storage.save_load_package_state(self.load_id, self.state)

    def on_resolved(self) -> None:
        self.state_save_lock = threading.Lock()
        self.state = self.storage.get_load_package_state(self.load_id)


def load_package_state() -> TLoadPackage:
    """Get full load package state present in current context. Across all threads this will be the same in memory dict."""
    container = Container()
    # get injected state if present. injected load package state is typically "managed" so changes will be persisted
    # if you need to save the load package state during a load, you need to call commit_load_package_state
    try:
        state_ctx = container[LoadPackageStateInjectableContext]
    except ContextDefaultCannotBeCreated:
        raise CurrentLoadPackageStateNotAvailable()
    return TLoadPackage(state=state_ctx.state, load_id=state_ctx.load_id)


def commit_load_package_state() -> None:
    """Commit load package state present in current context. This is thread safe."""
    container = Container()
    try:
        state_ctx = container[LoadPackageStateInjectableContext]
    except ContextDefaultCannotBeCreated:
        raise CurrentLoadPackageStateNotAvailable()
    state_ctx.commit()


def destination_state() -> DictStrAny:
    """Get segment of load package state that is specific to the current destination."""
    lp = load_package_state()
    return lp["state"].setdefault("destination_state", {})


def clear_destination_state(commit: bool = True) -> None:
    """Clear segment of load package state that is specific to the current destination. Optionally commit to load package."""
    lp = load_package_state()
    lp["state"].pop("destination_state", None)
    if commit:
        commit_load_package_state()



import dataclasses
import sys
from typing import Any, ClassVar, Final, List, Tuple, Union, Dict, Optional

from dlt.common.json import json
from dlt.common.pendulum import pendulum
from dlt.common.configuration.specs.api_credentials import OAuth2Credentials
from dlt.common.configuration.specs.exceptions import (
    InvalidGoogleNativeCredentialsType,
    InvalidGoogleOauth2Json,
    InvalidGoogleServicesJson,
    NativeValueError,
    OAuth2ScopesRequired,
    UnsupportedAuthenticationMethodException,
)
from dlt.common.configuration.specs.mixins import WithObjectStoreRsCredentials, WithPyicebergConfig
from dlt.common.exceptions import MissingDependencyException
from dlt.common.typing import DictStrAny, TSecretStrValue, StrAny, Self
from dlt.common.configuration.specs.base_configuration import (
    CredentialsConfiguration,
    CredentialsWithDefault,
    configspec,
)
from dlt.common.utils import is_interactive


# GCS scope required for OAuth2 token requests
GCS_SCOPE = "https://www.googleapis.com/auth/devstorage.read_write"


def _get_pyiceberg_fileio_config(credentials: Any, project_id: Optional[str]) -> Dict[str, Any]:
    """Get pyiceberg FileIO config from Google credentials.

    This refreshes the credentials to get a valid OAuth2 token and returns
    the configuration dict that pyiceberg expects for GCS access.
    """
    from google.auth.transport.requests import Request

    if not credentials.token:
        # For service account credentials, we need to add scopes before refresh
        # The credentials.requires_scopes check handles both service account and OAuth credentials
        if credentials.requires_scopes:
            credentials = credentials.with_scopes([GCS_SCOPE])
        credentials.refresh(Request())

    config: Dict[str, Any] = {}
    config["gcs.project-id"] = project_id
    config["gcs.oauth2.token"] = credentials.token
    if credentials.expiry:
        expiry_ms = int(credentials.expiry.timestamp() * 1000)
        config["gcs.oauth2.token-expires-at"] = str(expiry_ms)

    return config


@configspec
class GcpCredentials(CredentialsConfiguration, WithObjectStoreRsCredentials):
    token_uri: Final[str] = dataclasses.field(
        default="https://oauth2.googleapis.com/token", init=False, repr=False, compare=False
    )
    auth_uri: Final[str] = dataclasses.field(
        default="https://accounts.google.com/o/oauth2/auth", init=False, repr=False, compare=False
    )

    project_id: str = None

    def parse_native_representation(self, native_value: Any) -> None:
        if not isinstance(native_value, str):
            raise InvalidGoogleNativeCredentialsType(self.__class__, native_value)

    def to_native_representation(self) -> str:
        return json.dumps(dict(self))

    def to_native_credentials(self) -> Any:
        """Returns respective native credentials for service account or oauth2 that can be passed to google clients"""
        pass

    def _from_info_dict(self, info: StrAny) -> None:
        self.update(info)

    def __str__(self) -> str:
        return f"{self.project_id}"

    def to_gcs_credentials(self) -> Dict[str, Any]:
        """
        Dict of keyword arguments that can be passed to gcsfs.
        Delegates default GCS credential handling to gcsfs.
        """
        return {
            "project": self.project_id,
            "token": (
                None
                if isinstance(self, CredentialsWithDefault) and self.has_default_credentials()
                else dict(self)
            ),
        }

    def to_object_store_rs_credentials(self) -> Dict[str, str]:
        """
        Dict of keyword arguments that can be passed to `object_store` Rust crate.
        Delegates default GCS credential handling to `object_store` Rust crate.
        """
        if isinstance(self, CredentialsWithDefault) and self.has_default_credentials():
            return {}
        return {"service_account_key": json.dumps(dict(self))}


@configspec
class GcpServiceAccountCredentialsWithoutDefaults(GcpCredentials, WithPyicebergConfig):
    private_key: TSecretStrValue = None
    private_key_id: Optional[str] = None
    client_email: str = None
    type: Final[str] = dataclasses.field(  # noqa: A003
        default="service_account", init=False, repr=False, compare=False
    )

    def parse_native_representation(self, native_value: Any) -> None:
        """Accepts ServiceAccountCredentials as native value. In other case reverts to serialized services.json"""
        service_dict: DictStrAny = None
        try:
            from google.oauth2.service_account import Credentials as ServiceAccountCredentials

            if isinstance(native_value, ServiceAccountCredentials):
                # extract credentials
                service_dict = {
                    "project_id": native_value.project_id,
                    "client_email": native_value.service_account_email,
                    "private_key": native_value,  # keep native credentials in private key
                }
                self.__is_resolved__ = True
        except ImportError:
            pass

        if service_dict is None:
            # check if type is str
            GcpCredentials.parse_native_representation(self, native_value)
            # if not instance of service account credentials then check type and try to parse native value
            try:
                service_dict = json.loads(native_value)
            except Exception:
                raise InvalidGoogleServicesJson(self.__class__, native_value)

        self._from_info_dict(service_dict)

    def on_resolved(self) -> None:
        if self.private_key and self.private_key[-1] != "\n":
            # must end with new line, otherwise won't be parsed by Crypto
            self.private_key = self.private_key + "\n"

    def to_native_credentials(self) -> Any:
        """Returns google.oauth2.service_account.Credentials"""
        from google.oauth2.service_account import Credentials as ServiceAccountCredentials

        if isinstance(self.private_key, ServiceAccountCredentials):
            # private key holds the native instance if it was passed to parse_native_representation
            return self.private_key
        else:
            return ServiceAccountCredentials.from_service_account_info(self)

    def to_pyiceberg_fileio_config(self) -> Dict[str, Any]:
        return _get_pyiceberg_fileio_config(self.to_native_credentials(), self.project_id)

    @classmethod
    def from_pyiceberg_fileio_config(cls, file_io: Dict[str, Any]) -> Self:
        raise UnsupportedAuthenticationMethodException(
            "Service Account authentication not supported with `iceberg` table format. Use OAuth"
            " authentication instead."
        )

    def __str__(self) -> str:
        return f"{self.client_email}@{self.project_id}"


@configspec
class GcpOAuthCredentialsWithoutDefaults(GcpCredentials, OAuth2Credentials, WithPyicebergConfig):
    # only desktop app supported
    refresh_token: TSecretStrValue = None
    client_type: Final[str] = dataclasses.field(
        default="installed", init=False, repr=False, compare=False
    )

    def parse_native_representation(self, native_value: Any) -> None:
        """Accepts Google OAuth2 credentials as native value. In other case reverts to serialized oauth client secret json"""
        oauth_dict: DictStrAny = None
        try:
            from google.oauth2.credentials import Credentials as GoogleOAuth2Credentials

            if isinstance(native_value, GoogleOAuth2Credentials):
                # extract credentials, project id may not be present
                oauth_dict = {
                    "project_id": native_value.quota_project_id or "",
                    "client_id": native_value.client_id,
                    "client_secret": native_value.client_secret,
                    "refresh_token": native_value.refresh_token,
                    "scopes": native_value.scopes,
                    "token": native_value.token,
                }
                # if token is present, we are logged in
                self.__is_resolved__ = native_value.token is not None
        except ImportError:
            pass

        if oauth_dict is None:
            # check if type is str
            GcpCredentials.parse_native_representation(self, native_value)
            # if not instance of oauth2 credentials try to parse native value
            try:
                oauth_dict = json.loads(native_value)
                # if there's single element in the dict, this is probably "installed" or "web" etc. (app type)
                if len(oauth_dict) == 1:
                    oauth_dict = next(iter(oauth_dict.values()))
            except Exception:
                raise InvalidGoogleOauth2Json(self.__class__, native_value)
        self._from_info_dict(oauth_dict)

    def to_native_representation(self) -> str:
        return json.dumps(self._info_dict())

    def to_object_store_rs_credentials(self) -> Dict[str, str]:
        raise UnsupportedAuthenticationMethodException(
            "OAuth authentication not supported with `delta` table format. Use Service Account or"
            " Application Default Credentials authentication instead."
        )

    def to_pyiceberg_fileio_config(self) -> Dict[str, Any]:
        # do auth only if token is not yet present
        if not self.token:
            self.auth()
        return {
            "gcs.project-id": self.project_id,
            "gcs.oauth2.token": self.token,
            "gcs.oauth2.token-expires-at": (pendulum.now().timestamp() + 60) * 1000,
        }

    @classmethod
    def from_pyiceberg_fileio_config(cls, file_io: Dict[str, Any]) -> Self:
        credentials: Self = cls()
        credentials.project_id = file_io.get("gcs.project-id")
        credentials.token = file_io.get("gcs.oauth2.token")
        # if token and project are set, credentials are resolved
        credentials.on_partial()
        return credentials

    def auth(self, scopes: Union[str, List[str]] = None, redirect_url: str = None) -> None:
        if not self.refresh_token:
            self.add_scopes(scopes)
            if not self.scopes:
                raise OAuth2ScopesRequired(self.__class__)
            assert (
                sys.stdin.isatty() or is_interactive()
            ), "Must have a tty or interactive mode for web flow"
            self.refresh_token, self.token = self._get_refresh_token(
                redirect_url or "http://localhost"
            )
        else:
            # if scopes or redirect_url:
            #     logger.warning("Please note that scopes and redirect_url are ignored when getting access token")
            self.token = self._get_access_token()

    def on_partial(self) -> None:
        """Allows for an empty refresh token if the session is interactive or tty is attached"""
        if sys.stdin.isatty() or is_interactive():
            self.refresh_token = ""
            # still partial - raise
            if not self.is_partial():
                self.resolve()
            self.refresh_token = None
        # if token and project are set, credentials are resolved
        if self.token and self.project_id:
            self.refresh_token = ""
            self.client_id = ""
            self.resolve()

    def _get_access_token(self) -> str:
        try:
            from requests_oauthlib import OAuth2Session
        except ModuleNotFoundError:
            raise MissingDependencyException("GcpOAuthCredentials", ["requests_oauthlib"])

        google = OAuth2Session(client_id=self.client_id, scope=self.scopes)
        extra = {"client_id": self.client_id, "client_secret": self.client_secret}
        token: str = google.refresh_token(
            token_url=self.token_uri, refresh_token=self.refresh_token, **extra
        )["access_token"]
        return token

    def _get_refresh_token(self, redirect_url: str) -> Tuple[str, str]:
        try:
            from google_auth_oauthlib.flow import InstalledAppFlow
        except ModuleNotFoundError:
            raise MissingDependencyException("GcpOAuthCredentials", ["google-auth-oauthlib"])
        flow = InstalledAppFlow.from_client_config(self._installed_dict(redirect_url), self.scopes)
        credentials = flow.run_local_server(port=0)
        return credentials.refresh_token, credentials.token

    def to_native_credentials(self) -> Any:
        """Returns google.oauth2.credentials.Credentials"""
        try:
            from google.oauth2.credentials import Credentials as GoogleOAuth2Credentials
        except ModuleNotFoundError:
            raise MissingDependencyException("GcpOAuthCredentials", ["google-auth-oauthlib"])

        credentials = GoogleOAuth2Credentials.from_authorized_user_info(info=dict(self))
        return credentials

    def _installed_dict(self, redirect_url: str = "http://localhost") -> StrAny:
        installed_dict = {self.client_type: self._info_dict()}

        if redirect_url:
            installed_dict[self.client_type]["redirect_uris"] = [redirect_url]
        return installed_dict

    def _info_dict(self) -> DictStrAny:
        info_dict = dict(self)
        # for desktop app
        info_dict["redirect_uris"] = ["http://localhost"]
        return info_dict

    def __str__(self) -> str:
        return f"{self.client_id}@{self.project_id}"


@configspec
class GcpDefaultCredentials(CredentialsWithDefault, GcpCredentials):
    _LAST_FAILED_DEFAULT: ClassVar[float] = 0.0

    def parse_native_representation(self, native_value: Any) -> None:
        """Accepts google credentials as native value"""
        try:
            from google.auth.credentials import Credentials as GoogleCredentials

            if isinstance(native_value, GoogleCredentials):
                self.project_id = self.project_id or native_value.quota_project_id
                self._set_default_credentials(native_value)
                # is resolved
                self.__is_resolved__ = True
                return
        except ImportError:
            pass
        raise NativeValueError(
            self.__class__, native_value, "Default Google Credentials not present"
        )

    @staticmethod
    def _get_default_credentials(retry_timeout_s: float = 600.0) -> Tuple[Any, str]:
        now = pendulum.now().timestamp()
        if now - GcpDefaultCredentials._LAST_FAILED_DEFAULT < retry_timeout_s:
            return None, None

        from google.auth import default as default_credentials
        from google.auth.exceptions import DefaultCredentialsError

        try:
            return default_credentials()  # type: ignore
        except DefaultCredentialsError:
            # prevent exception
            GcpDefaultCredentials._LAST_FAILED_DEFAULT = now
            return None, None

    def on_partial(self) -> None:
        """Looks for default google credentials and resolves configuration if found. Otherwise continues as partial"""
        try:
            # if config is missing check if credentials can be obtained from defaults
            default, project_id = GcpDefaultCredentials._get_default_credentials()
            if default is None:
                return
            # set the project id - it needs to be known by the client
            self.project_id = self.project_id or project_id or default.quota_project_id
            self._set_default_credentials(default)
            self.resolve()
        except ImportError:
            # raise the exception that caused partial (typically missing config fields)
            pass

    def to_native_credentials(self) -> Any:
        if self.has_default_credentials():
            return self.default_credentials()
        else:
            return super().to_native_credentials()


@configspec
class GcpServiceAccountCredentials(
    GcpDefaultCredentials, GcpServiceAccountCredentialsWithoutDefaults
):
    def parse_native_representation(self, native_value: Any) -> None:
        try:
            GcpDefaultCredentials.parse_native_representation(self, native_value)
        except NativeValueError:
            pass
        GcpServiceAccountCredentialsWithoutDefaults.parse_native_representation(self, native_value)

    def to_pyiceberg_fileio_config(self) -> Dict[str, Any]:
        if self.has_default_credentials():
            return _get_pyiceberg_fileio_config(self.default_credentials(), self.project_id)
        else:
            return GcpServiceAccountCredentialsWithoutDefaults.to_pyiceberg_fileio_config(self)


@configspec
class GcpOAuthCredentials(GcpDefaultCredentials, GcpOAuthCredentialsWithoutDefaults):
    def parse_native_representation(self, native_value: Any) -> None:
        try:
            GcpDefaultCredentials.parse_native_representation(self, native_value)
        except NativeValueError:
            pass
        GcpOAuthCredentialsWithoutDefaults.parse_native_representation(self, native_value)

    def to_pyiceberg_fileio_config(self) -> Dict[str, Any]:
        if self.has_default_credentials():
            return _get_pyiceberg_fileio_config(self.default_credentials(), self.project_id)
        else:
            return GcpOAuthCredentialsWithoutDefaults.to_pyiceberg_fileio_config(self)



import os
from typing import Optional
from urllib.parse import urlparse, urlunparse

from dlt.common.configuration.specs import (
    CredentialsConfiguration,
    AwsCredentialsWithoutDefaults,
    AzureCredentialsWithoutDefaults,
)
from dlt.common.destination.configuration import CsvFormatConfiguration
from dlt.common.storages import FilesystemConfiguration
from dlt.common.storages.configuration import ensure_canonical_az_url
from dlt.common.storages.fsspec_filesystem import (
    AZURE_BLOB_STORAGE_PROTOCOLS,
    S3_PROTOCOLS,
)

from dlt.common.typing import TLoaderFileFormat
from dlt.destinations.exceptions import LoadJobTerminalException


SNOWFLAKE_SESSION_TOKEN_PATH = "/snowflake/session/token"


def ensure_snowflake_azure_url(
    file_url: str, account_name: str = None, account_host: str = None
) -> str:
    # Explicit azure credentials are needed to load from bucket without a named stage
    if not account_host and account_name:
        account_host = f"{account_name}.blob.core.windows.net"
    # get canonical url first to convert it into snowflake form
    canonical_url = ensure_canonical_az_url(
        file_url,
        "azure",
        account_name,
        account_host,
    )
    parsed_file_url = urlparse(canonical_url)
    return urlunparse(
        parsed_file_url._replace(
            path=f"/{parsed_file_url.username}{parsed_file_url.path}",
            netloc=parsed_file_url.hostname,
        )
    )


def _build_format_clause(format_clause: list[str]) -> str:
    joined = ", ".join(format_clause)
    return f"({joined})"


def gen_copy_sql(
    file_url: str,
    qualified_table_name: str,
    loader_file_format: TLoaderFileFormat,
    is_case_sensitive: bool,
    stage_name: Optional[str] = None,
    stage_bucket_url: Optional[str] = None,
    local_stage_file_path: Optional[str] = None,
    staging_credentials: Optional[CredentialsConfiguration] = None,
    csv_format: Optional[CsvFormatConfiguration] = None,
    use_vectorized_scanner: Optional[bool] = False,
) -> str:
    """
    Generates a Snowflake COPY command to load data from a file.

    Args:
        use_vectorized_scanner: Whether to use the vectorized scanner in COPY INTO
        file_url: URL of the file to load
        qualified_table_name: Fully qualified name of the target table
        loader_file_format: Format of the source file (jsonl, parquet, csv)
        is_case_sensitive: Whether column matching should be case-sensitive
        stage_name: Optional name of a predefined Snowflake stage
        stage_bucket_url: Optional URL of the bucket containing the file
        local_stage_file_path: Path to use for local files
        staging_credentials: Optional credentials for accessing cloud storage
        csv_format: Optional configuration for CSV format

    Returns:
        A SQL string containing the COPY command
    """
    # Determine file location type and get potentially modified URL
    parsed_file_url = urlparse(file_url)

    # Initialize clause components
    credentials_clause = ""
    files_clause = ""

    # Handle different file location types
    if parsed_file_url.scheme == "file" or FilesystemConfiguration.is_local_path(file_url):
        from_clause = f"FROM {local_stage_file_path}"

    elif stage_name:
        relative_url = parsed_file_url.path.lstrip("/")

        # If stage bucket URL has a path, remove it from the beginning of the relative URL because this is already specified
        # in the Snowflake stage location.
        if stage_bucket_url:
            parsed_bucket_url = urlparse(stage_bucket_url)
            stage_bucket_path = parsed_bucket_url.path.lstrip("/")
            if stage_bucket_path:
                relative_url = relative_url.removeprefix(stage_bucket_path)

        from_clause = f"FROM @{stage_name}"
        files_clause = f"FILES = ('{relative_url}')"

    elif parsed_file_url.scheme in S3_PROTOCOLS:
        if staging_credentials and isinstance(staging_credentials, AwsCredentialsWithoutDefaults):
            credentials_clause = (
                f"CREDENTIALS=(AWS_KEY_ID='{staging_credentials.aws_access_key_id}' "
                f"AWS_SECRET_KEY='{staging_credentials.aws_secret_access_key}')"
            )
            from_clause = f"FROM '{file_url}'"
        else:
            raise LoadJobTerminalException(
                file_url,
                f"Cannot load from S3 path `{file_url}` without either credentials or a stage"
                " name.",
            )

    elif parsed_file_url.scheme in AZURE_BLOB_STORAGE_PROTOCOLS:
        if staging_credentials and isinstance(staging_credentials, AzureCredentialsWithoutDefaults):
            credentials_clause = (
                f"CREDENTIALS=(AZURE_SAS_TOKEN='?{staging_credentials.azure_storage_sas_token}')"
            )
            converted_az_url = ensure_snowflake_azure_url(
                file_url,
                staging_credentials.azure_storage_account_name,
                staging_credentials.azure_account_host,
            )
            from_clause = f"FROM '{converted_az_url}'"
        else:
            raise LoadJobTerminalException(
                file_url,
                f"Cannot load from Azure path `{file_url}` without either credentials or a stage"
                " name.",
            )
    else:
        raise LoadJobTerminalException(
            file_url,
            f"Cannot load from bucket path `{file_url}` without a stage name. See "
            "https://dlthub.com/docs/dlt-ecosystem/destinations/snowflake for "
            "instructions on setting up the `stage_name`",
        )

    # Generate file format clause
    if is_case_sensitive:
        column_match_clause = "MATCH_BY_COLUMN_NAME='CASE_SENSITIVE'"
    else:
        column_match_clause = "MATCH_BY_COLUMN_NAME='CASE_INSENSITIVE'"

    if loader_file_format == "jsonl":
        on_error_clause = ""
        format_opts = [
            "TYPE = 'JSON'",
            "BINARY_FORMAT = 'BASE64'",
        ]
    elif loader_file_format == "parquet":
        format_opts = [
            "TYPE = 'PARQUET'",
            "BINARY_AS_TEXT = FALSE",
            "USE_LOGICAL_TYPE = TRUE",
        ]
        if use_vectorized_scanner:
            format_opts.append("USE_VECTORIZED_SCANNER = TRUE")
            on_error_clause = "ON_ERROR = ABORT_STATEMENT"
        else:
            on_error_clause = ""
    elif loader_file_format == "csv":
        csv_config = csv_format or CsvFormatConfiguration()
        if not csv_config.include_header:
            column_match_clause = ""
        if csv_config.on_error_continue:
            on_error_clause = "ON_ERROR = CONTINUE"
        else:
            on_error_clause = ""
        format_opts = [
            "TYPE = 'CSV'",
            "BINARY_FORMAT = 'UTF-8'",
            f"PARSE_HEADER = {csv_config.include_header}",
            "FIELD_OPTIONALLY_ENCLOSED_BY = '\"'",
            "NULL_IF = ('')",
            "ERROR_ON_COLUMN_COUNT_MISMATCH = FALSE",
            f"FIELD_DELIMITER = '{csv_config.delimiter}'",
            f"ENCODING = '{csv_config.encoding}'",
        ]
    else:
        raise ValueError(f"`{loader_file_format=:}` not supported for Snowflake COPY command.")

    source_format = _build_format_clause(format_opts)

    # Construct the final SQL statement
    return f"""COPY INTO {qualified_table_name}
        {from_clause}
        {files_clause}
        {credentials_clause}
        FILE_FORMAT = {source_format}
        {column_match_clause}
        {on_error_clause}
    """


def snowflake_session_token_available() -> bool:
    """Whether Snowflake-provided OAuth token is available in environment.

    Reference: https://docs.snowflake.com/en/developer-guide/snowpark-container-services/additional-considerations-services-jobs#connecting-with-a-snowflake-provided-oauth-token
    """
    return (
        os.path.exists(SNOWFLAKE_SESSION_TOKEN_PATH)
        and "SNOWFLAKE_ACCOUNT" in os.environ
        and "SNOWFLAKE_HOST" in os.environ
    )


def read_snowflake_session_token() -> str:
    """
    Read the session token supplied automatically by Snowflake. These tokens
    are short lived and should always be read right before creating any new connection.
    """
    with open(SNOWFLAKE_SESSION_TOKEN_PATH, "r", encoding="utf-8") as f:
        return f.read()



import platform
import os
from dlt.common import logger

if platform.python_implementation() == "PyPy":
    import psycopg2cffi as psycopg2

    # from psycopg2cffi.sql import SQL, Composed
else:
    import psycopg2

    # from psycopg2.sql import SQL, Composed

from typing import Dict, List, Optional, Sequence

from dlt.common.destination.client import (
    FollowupJobRequest,
    CredentialsConfiguration,
    PreparedTableSchema,
    SupportsStagingDestination,
    LoadJob,
)
from dlt.common.destination.capabilities import DestinationCapabilitiesContext
from dlt.common.schema import TColumnSchema, TColumnHint, Schema
from dlt.common.schema.utils import table_schema_has_type
from dlt.common.schema.typing import TColumnType
from dlt.common.configuration.specs import AwsCredentialsWithoutDefaults

from dlt.destinations.insert_job_client import InsertValuesJobClient
from dlt.destinations.sql_jobs import SqlMergeFollowupJob
from dlt.destinations.exceptions import DatabaseTerminalException, DatabaseException
from dlt.destinations.job_client_impl import CopyRemoteFileLoadJob
from dlt.destinations.impl.postgres.sql_client import Psycopg2SqlClient
from dlt.destinations.impl.redshift.configuration import RedshiftClientConfiguration
from dlt.destinations.job_impl import ReferenceFollowupJobRequest
from dlt.destinations.path_utils import get_file_format_and_compression

HINT_TO_REDSHIFT_ATTR: Dict[TColumnHint, str] = {
    "cluster": "DISTKEY",
    # it is better to not enforce constraints in redshift
    # "primary_key": "PRIMARY KEY",
    "sort": "SORTKEY",
}


class RedshiftSqlClient(Psycopg2SqlClient):
    def has_dataset(self) -> bool:
        # In Redshift, the 'public' schema always exists but may not be
        # returned by INFORMATION_SCHEMA.SCHEMATA query, so we handle it as a special case
        if self.dataset_name.lower() == "public":
            return True

        all_schemas_view_name = "svv_redshift_schemas"

        db_params = []
        _, schema_name, _ = self._get_information_schema_components()
        query = f"SELECT 1 FROM {all_schemas_view_name} WHERE schema_name = %s"
        db_params.append(schema_name)
        try:
            rows = self.execute_sql(query, *db_params)
            return len(rows) > 0
        except DatabaseException:
            # fallback to INFORMATION_SCHEMA.SCHEMATA if svv_redshift_schemas is not available
            return super().has_dataset()

    @staticmethod
    def _maybe_make_terminal_exception_from_data_error(
        pg_ex: psycopg2.DataError,
    ) -> Optional[Exception]:
        if "Cannot insert a NULL value into column" in pg_ex.pgerror:
            # NULL violations is internal error, probably a redshift thing
            return DatabaseTerminalException(pg_ex)
        if "Numeric data overflow" in pg_ex.pgerror:
            return DatabaseTerminalException(pg_ex)
        if "Precision exceeds maximum" in pg_ex.pgerror:
            return DatabaseTerminalException(pg_ex)
        return None


class RedshiftCopyFileLoadJob(CopyRemoteFileLoadJob):
    def __init__(
        self,
        file_path: str,
        staging_credentials: Optional[CredentialsConfiguration] = None,
        staging_iam_role: str = None,
    ) -> None:
        super().__init__(file_path, staging_credentials)
        self._staging_iam_role = staging_iam_role
        self._job_client: "RedshiftClient" = None

    def run(self) -> None:
        self._sql_client = self._job_client.sql_client
        # we assume s3 credentials where provided for the staging
        aws_region = (
            self._staging_credentials.get("region_name", "") if self._staging_credentials else ""
        )
        region = f"region '{aws_region}'" if aws_region else ""
        credentials = ""
        if self._staging_iam_role:
            credentials = f"IAM_ROLE '{self._staging_iam_role}'"
        elif self._staging_credentials and isinstance(
            self._staging_credentials, AwsCredentialsWithoutDefaults
        ):
            aws_access_key = self._staging_credentials.aws_access_key_id
            aws_secret_key = self._staging_credentials.aws_secret_access_key
            aws_session_token = self._staging_credentials.aws_session_token
            credentials = (
                "CREDENTIALS"
                f" 'aws_access_key_id={aws_access_key};aws_secret_access_key={aws_secret_key}'"
            )
            if aws_session_token:
                credentials = credentials[:-1] + f";token={aws_session_token}'"
        # get format
        file_format, is_compressed = get_file_format_and_compression(self._bucket_path)
        file_type = ""
        dateformat = ""
        compression = ""
        if file_format == "jsonl":
            file_type = "FORMAT AS JSON 'auto'"
            dateformat = "dateformat 'auto' timeformat 'auto'"
            compression = "" if not is_compressed else "GZIP"
        elif file_format == "parquet":
            # Redshift doesn't support copying across regions for columnar data formats
            # https://docs.aws.amazon.com/redshift/latest/dg/copy-usage_notes-copy-from-columnar.html  # noqa: E501
            if region:
                logger.warning(
                    "Redshift doesn't support copying Parquet files across regions. The region"
                    " parameter will be ignored."
                )
                region = ""

            file_type = "PARQUET"
            # if table contains json types then SUPER field will be used.
            # https://docs.aws.amazon.com/redshift/latest/dg/ingest-super.html
            if table_schema_has_type(self._load_table, "json"):
                file_type += " SERIALIZETOJSON"
        else:
            raise ValueError(f"Unsupported file type `{file_format}` for Redshift.")

        with self._sql_client.begin_transaction():
            # TODO: if we ever support csv here remember to add column names to COPY
            self._sql_client.execute_sql(f"""
                COPY {self._sql_client.make_qualified_table_name(self.load_table_name)}
                FROM '{self._bucket_path}'
                {file_type}
                {dateformat}
                {compression}
                {credentials}
                {region} MAXERROR 0;""")


class RedshiftMergeJob(SqlMergeFollowupJob):
    @classmethod
    def gen_key_table_clauses(
        cls,
        root_table_name: str,
        staging_root_table_name: str,
        key_clauses: Sequence[str],
        for_delete: bool,
    ) -> List[str]:
        """Generate sql clauses that may be used to select or delete rows in root table of destination dataset

        A list of clauses may be returned for engines that do not support OR in subqueries. Like BigQuery
        """
        if for_delete:
            return [
                f"FROM {root_table_name} WHERE EXISTS (SELECT 1 FROM"
                f" {staging_root_table_name} WHERE"
                f" {' OR '.join([c.format(d=root_table_name,s=staging_root_table_name) for c in key_clauses])})"
            ]
        return SqlMergeFollowupJob.gen_key_table_clauses(
            root_table_name, staging_root_table_name, key_clauses, for_delete
        )


class RedshiftClient(InsertValuesJobClient, SupportsStagingDestination):
    def __init__(
        self,
        schema: Schema,
        config: RedshiftClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        dataset_name, staging_dataset_name = InsertValuesJobClient.create_dataset_names(
            schema, config
        )
        sql_client = RedshiftSqlClient(
            dataset_name,
            staging_dataset_name,
            config.credentials,
            capabilities,
        )
        super().__init__(schema, config, sql_client)
        self.active_hints = HINT_TO_REDSHIFT_ATTR
        self.sql_client = sql_client
        self.config: RedshiftClientConfiguration = config
        self.type_mapper = self.capabilities.get_type_mapper()

    def _create_merge_followup_jobs(
        self, table_chain: Sequence[PreparedTableSchema]
    ) -> List[FollowupJobRequest]:
        return [RedshiftMergeJob.from_table_chain(table_chain, self.sql_client)]

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        """Starts SqlLoadJob for files ending with .sql or returns None to let derived classes to handle their specific jobs"""
        job = super().create_load_job(table, file_path, load_id, restore)
        if not job:
            assert ReferenceFollowupJobRequest.is_reference_job(
                file_path
            ), "Redshift must use staging to load files"
            job = RedshiftCopyFileLoadJob(
                file_path,
                staging_credentials=self.config.staging_config.credentials,
                staging_iam_role=self.config.staging_iam_role,
            )
        return job

    def _from_db_type(
        self, pq_t: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        return self.type_mapper.from_destination_type(pq_t, precision, scale)

    def should_truncate_table_before_load_on_staging_destination(self, table_name: str) -> bool:
        return self.config.truncate_tables_on_staging_destination_before_load



import marimo

__generated_with = "0.19.2"
app = marimo.App(width="medium")

with app.setup:
    import marimo as mo
    from dlt.helpers.marimo.utils import _load_file

    pipeline_path = None


@app.cell
def pipeline_browser():
    mo.stop(pipeline_path is None)
    pipeline_details = mo.ui.file_browser(
        pipeline_path,
        selection_mode="file",
        multiple=False,
        restrict_navigation=True,
        label="Select file to load",
    )
    pipeline_details
    return (pipeline_details,)


@app.cell
def file_viewer(pipeline_details):
    if pipeline_details.value:
        obj = _load_file(pipeline_details.value[0].path)
    else:
        obj = None

    obj
    return


if __name__ == "__main__":
    app.run()
